import subprocess
import os
import re
import csv


class MalletError(Exception):
    pass


class MalletTopicModel(object):
    def __init__(self, mallet_dir, memory=1):
        '''
        Constructor
        mallet_dir : str : file path of Mallet-3 directory
        memory : int, float : maximum gigabytes of memory to allocate to Mallet
        '''
        self.memory = memory
        # self.mallet_dir = mallet_dir if mallet_dir.endswith('/') else mallet_dir + '/'
        self.mallet_dir = mallet_dir

        self.set_memory()
        self.set_output_dir()

    def set_memory(self):
        '''
        Sets the memory limit by editing a line in a batch file
        Default from constructor is 1gb
        '''
        # reads
        full_path = os.path.join(self.mallet_dir, r'bin\mallet.bat' if os.name == 'nt' else 'bin/mallet')
        with open(full_path, 'r') as file:
            filedata = file.read()
        # replaces memory line
        filedata = re.sub('set MALLET_MEMORY=.*', 'set MALLET_MEMORY=' + str(self.memory) + 'g', filedata)
        # writes
        with open(full_path, 'w') as file:
            file.write(filedata)

    def set_output_dir(self):
        '''
        Creates (if not already present) an output folder in the Mallet directory
        Sets stdout file path
        '''
        if not os.path.isdir(self.mallet_dir + 'output'):
            os.mkdir(self.mallet_dir + 'output')

        outdir_full_path = os.path.join(self.mallet_dir, 'output')
        self.output_dir = outdir_full_path
        self.out_full_path = os.path.join(self.output_dir, 'output.txt')
        self.stdout = self.out_full_path

    def call(self, command):
        '''
        Executes a command within the Mallet directory
        Redirects all output to a txt file
        '''

        print(command)
        result = subprocess.call(command, cwd=self.mallet_dir, stderr=subprocess.STDOUT, stdout=open(self.stdout, 'w'),
                                 shell=True)

    def execute(self, command, args):
        '''
        Cannot determine success/failure of Mallet command execution from exit codes (0 or 1)
        Checks whether the file that was supposed to be generated by a command is actually generated
        If the file did not previously exist, checks if it exists after command execution
        Otherwise, checks whether the modified timestamp of the file has changed
        '''
        start = None
        for arg in args:
            # for train_topics, this will be either 'output-topic-keys' or 'output-doc-topics' (not 'topic-word-weights-file')
            # it should not matter which of the three are checked; if the command fails, neither should be generated
            if 'output' in arg:
                if os.path.isfile(args[arg]):
                    start = os.path.getmtime(args[arg])
                break
        self.call(command)
        stop = None
        if os.path.isfile(args[arg]):
            stop = os.path.getmtime(args[arg])
        if not stop or start == stop:
            raise MalletError('\n\n' + args[arg] + ' was not generated from the command:\n\n"' + ' '.join(command) +
                              '"\n\nCheck ' + self.out_full_path + ' for details.\n')

    def build_command(self, operation, kwargs):
        '''
        Converts the command arguments from dictionary to list format
        '''
        command = [r'bin\mallet.bat' if os.name == 'nt' else 'bin/mallet', operation]
        for key, value in kwargs.items():
            if isinstance(value, list):
                command.extend(['--' + key.replace('_', '-'), ' '.join(value)])
            else:
                command.extend(['--' + key.replace('_', '-'), value])
        return command

    def import_dir(self, **kwargs):
        '''
        Calls the import-dir Mallet command
        Overwrites some arguments
        '''
        kwargs[
            'keep_sequence'] = True  # topic modeling currently only supports feature sequences: use --keep-sequence option when importing data.
        full_path = os.path.join(self.output_dir, 'import.mallet')
        kwargs['output'] = full_path
        command = self.build_command(operation='import-dir', kwargs=kwargs)
        self.execute(command, kwargs)

    def import_file(self, **kwargs):
        '''
        Calls the import-file Mallet command
        Overwrites some arguments
        '''

        full_path = os.path.join(self.output_dir, 'import.mallet')
        kwargs['output'] = full_path

        kwargs['keep_sequence'] = str(True) if os.name == 'nt' else True
        # topic modeling currently only supports feature sequences: use --keep-sequence option when importing data.
        kwargs['label-as-features'] = str(True) if os.name == 'nt' else True
        command = self.build_command(operation='import-file', kwargs=kwargs)
        self.execute(command, kwargs)

    def train_topics(self, **kwargs):
        '''
        Calls the train-topics Mallet command
        Overwrites some arguments
        '''
        instance_input = os.path.join(self.output_dir, 'import.mallet')
        kwargs['input'] = instance_input
        self.word_weights_file = os.path.join(self.output_dir, 'topic_word_weights.txt')
        kwargs['topic-word-weights-file'] = self.word_weights_file
        self.topic_keys_file = os.path.join(self.output_dir, 'topic_keys.txt')
        kwargs['output-topic-keys'] = self.topic_keys_file  # these can also be derived from topic-word-weights-file
        self.doc_topics_file = os.path.join(self.output_dir, 'doc_topics.txt')
        kwargs['output-doc-topics'] = self.doc_topics_file

        self.parameter_file = os.path.join(self.output_dir, 'topic_parameter.txt')
        kwargs['parameter-filename'] = self.parameter_file

        kwargs['num-top-words'] = str(30) if os.name == 'nt' else 30
        kwargs['num-iterations'] = str(2000) if os.name == 'nt' else 2000
        kwargs['num-topics'] = str(10) if os.name == 'nt' else 10

        # sets the doc-topics-threshold to a very low number if not specified
        # output-doc-topics format changes when this parameter is not specified
        # see https://github.com/mimno/Mallet/issues/41
        if not kwargs.get('doc-topics-threshold', False):
            kwargs['doc-topics-threshold'] = str(0.0001) if os.name == 'nt' else 0.0001
        command = self.build_command(operation='train-topics', kwargs=kwargs)
        self.execute(command, kwargs)

        # parses the files generated by the command into python objects
        self.parse_topic_keys()
        self.parse_doc_topics()
        self.parse_word_weights()

    def parse_topic_keys(self):
        '''
        Parses the output-topic-keys txt file generated by train_topics into a python object
        {topic # (int): {dirichlet parameter: float, words: list}, ... }
        '''
        self.topic_keys = dict()
        with open(self.topic_keys_file, mode='r', encoding='utf8') as file:
            for line in file:
                data = re.split('\t| ',
                                line.strip())  # cols 0, 1, and 2 are tab delimited, the rest are space delimited
                self.topic_keys[int(data[0])] = {'dirichlet': float(data[1]), 'words': self.process_words(data[2:])}

    def parse_doc_topics(self):
        '''
        Parses the output-doc-topics txt file generated by train_topics into a python object
        {document # (int): {document name: e.g. file path (str), topics: {topic # (int): weight (float), ... }}, ... }
        '''
        self.doc_topics = dict()
        with open(self.doc_topics_file, mode='r', encoding='utf8') as file:
            next(file)
            for line in file:
                data = line.strip().split('\t')
                self.doc_topics[int(data[0])] = {'name': data[1], 'topics': self.process_topics(data[2:])}

    def parse_word_weights(self):
        '''
        Parses the topic-word-weights txt file generated by train_topics into a python object
        {topic # (int): {word (str): weight (float)}, ... }
        '''
        self.word_weights = dict()
        with open(self.word_weights_file, mode='r', encoding='utf8') as file:
            for line in file:
                data = line.strip().split('\t')
                try:
                    self.word_weights[int(data[0])][data[1]] = float(data[2])
                except KeyError:
                    self.word_weights[int(data[0])] = {data[1]: float(data[2])}

    @staticmethod
    def process_words(words):
        return [x.replace('_', ' ') for x in words]

    @staticmethod
    def process_topics(topics):
        topic_weights = dict()
        for i in range(0, len(topics), 2):
            topic_weights[int(topics[i])] = float(topics[i + 1])
        return topic_weights


if __name__ == '__main__':
    model = MalletTopicModel('C:\mallet')
    # model.import_file(input=r'C:\mallet\topic_input\dblp_sample.txt')
    model.import_file(input=r'C:\mallet\topic_input\sample_dmr_input.txt')
    model.train_topics()

    print(model.topic_keys)  # see output_topic_keys parameter in Train Topics documentation
    print(model.doc_topics)  # see output_doc_topics parameter in Train Topics documentation
    print(model.word_weights)  # see topic_word_weights_file parameter in Train Topics documentation



class BaseSynonymExtraction:
    IN_TYPE = [str]
    OUT_TYPE = [str]

class WordNetSynonymExtraction(BaseSynonymExtraction):
    def __init__(self):
        import nltk
        try:
            wnl = nltk.WordNetLemmatizer()
        except LookupError:
            nltk.download('wordnet')

    def __call__(self, *args, **kwargs):
        from nltk.corpus import wordnet
        # Creating a list
        synonyms = []
        for syn in wordnet.synsets(args):
            for lm in syn.lemmas():
                synonyms.append(lm.name())  # adding into synonyms
        print(set(synonyms))
        return synonyms


import sys
import os
import re
import pickle
from   optparse import OptionParser
import numpy as np
import tensorflow as tf

import treform.segmentation.lstmWordSegmentationModel as model
import treform.segmentation.wordSegmentationModelUtil as util

VERBOSE = 0

if __name__ == '__main__':
    VERBOSE = 1
    train_path = "D:\\python_workspace\\treform\\data\\korean_lstm_segmentation_train.txt"
    validation_path = "D:\\python_workspace\\treform\\data\\korean_lstm_segmentation_validation.txt"
    model_dir = "./model"

    training_iters = 100
    if not os.path.isdir(model_dir) :
        os.makedirs(model_dir)

    # config
    n_steps = 30                    # time steps
    padd = '\t'                     # special padding chracter
    char_dic = util.build_dictionary(train_path, padd)
    n_input = len(char_dic)         # input dimension, vocab size
    n_hidden = 8                    # hidden layer size
    n_classes = 2                   # output classes,  space or not
    vocab_size = n_input
    '''
    util.test_next_batch(train_path, char_dic, vocab_size, n_steps, padd)
    '''
    x = tf.placeholder(tf.float32, [None, n_steps, n_input])
    y_ = tf.placeholder(tf.int32, [None, n_steps])
    early_stop = tf.placeholder(tf.int32)

    # LSTM layer
    # 2 x n_hidden = state_size = (hidden state + cell state)
    istate = tf.placeholder(tf.float32, [None, 2*n_hidden])
    weights = {
        'hidden' : model.weight_variable([n_input, n_hidden]),
        'out' : model.weight_variable([n_hidden, n_classes])
    }
    biases = {
        'hidden' : model.bias_variable([n_hidden]),
        'out': model.bias_variable([n_classes])
    }

    # training
    y = model.RNN(x, istate, weights, biases, n_hidden, n_steps, n_input, early_stop)

    batch_size = 1
    learning_rate = 0.01
    logits = tf.reshape(tf.concat(y, 1), [-1, n_classes])
    targets = y_
    seq_weights = tf.ones([n_steps * batch_size])
    loss = tf.contrib.legacy_seq2seq.sequence_loss_by_example([logits], [targets], [seq_weights])
    cost = tf.reduce_sum(loss) / batch_size
    optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cost)

    correct_pred = tf.equal(tf.argmax(logits,1), tf.cast(y_, tf.int64))
    accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

    NUM_THREADS = 1
    config = tf.ConfigProto(intra_op_parallelism_threads=NUM_THREADS,
            inter_op_parallelism_threads=NUM_THREADS,
            log_device_placement=False)
    sess = tf.Session(config=config)
    init = tf.global_variables_initializer()
    sess.run(init)
    saver = tf.train.Saver() # save all variables
    checkpoint_dir = model_dir
    checkpoint_file = 'segm.ckpt'

    sys.stderr.write('save dic\n')
    dic_path = model_dir + '/' + 'dic.pickle'
    with open(dic_path, 'wb') as handle :
        pickle.dump(char_dic, handle)

    validation_data = util.get_validation_data(validation_path, char_dic, vocab_size, n_steps, padd)

    seq = 0
    while seq < training_iters :
        c_istate = np.zeros((batch_size, 2*n_hidden))
        i = 0
        fid = util.open_file(train_path, 'r')
        for line in fid :
            line = line.strip()
            if line == "" : continue
            sentence = util.snorm(line)
            pos = 0
            while pos != -1 :
                batch_xs, batch_ys, next_pos, count = util.next_batch(sentence, pos, char_dic, vocab_size, n_steps, padd)
                '''
                print 'window : ' + sentence[pos:pos+n_steps].encode('utf-8')
                print 'count : ' + str(count)
                print 'next_pos : ' + str(next_pos)
                print batch_ys
                print batch_xs
                '''
                feed={x: batch_xs, y_: batch_ys, istate: c_istate, early_stop:count}
                sess.run(optimizer, feed_dict=feed)
                pos = next_pos
            sys.stderr.write('%s th sentence for %s th iterations ... done\n' % (i, seq))
            i += 1
        util.close_file(fid)
        # validation
        if seq % 1 == 0 :
            validation_cost = 0
            validation_accuracy = 0
            for validation_xs, validation_ys, count in validation_data :
                feed={x: validation_xs, y_: validation_ys, istate: c_istate, early_stop:count}
                validation_cost += sess.run(cost, feed_dict=feed)
                validation_accuracy += sess.run(accuracy, feed_dict=feed)
            validation_cost /= len(validation_data)
            validation_accuracy /= len(validation_data)
            sys.stderr.write('iterations : %s' % (seq) + ',' + 'validation cost : %s' % validation_cost + ',' + 'validation accuracy : %s\n' % (validation_accuracy))
            sys.stderr.write('save model\n')
            saver.save(sess, checkpoint_dir + '/' + checkpoint_file)
        seq += 1

    sys.stderr.write('save model(final)\n')
    saver.save(sess, checkpoint_dir + '/' + checkpoint_file)
    sys.stderr.write('end of training\n')
    sess.close()




''' str => list(str) '''

class BaseTokenizer:
    IN_TYPE = [str]
    OUT_TYPE = [list, str]

# [English]
class Tweet(BaseTokenizer):
    def __init__(self):
        import nltk.tokenize
        self.inst = nltk.tokenize.TweetTokenizer()

    def __call__(self, *args, **kwargs):
        return self.inst.tokenize(*args)

class Whitespace(BaseTokenizer):
    def __init__(self):
        import nltk.tokenize
        self.inst = nltk.tokenize.WhitespaceTokenizer()

    def __call__(self, *args, **kwargs):
        return self.inst.tokenize(*args)

class Word(BaseTokenizer):
    def __init__(self):
        import nltk.tokenize
        self.inst = nltk.tokenize.word_tokenize

    def __call__(self, *args, **kwargs):
        #print(str(self.inst(*args)))
        return self.inst(*args)

class WordPos(BaseTokenizer):
    def __init__(self):
        import nltk
        self.inst = nltk
        self.OUT_TYPE = [list, tuple]

    def __call__(self, *args, **kwargs):
        tokens = self.inst.word_tokenize(*args)

        return self.inst.pos_tag(tokens)

# [Korean]
class Komoran(BaseTokenizer):
    def __init__(self,userdic=None):
        from konlpy.tag import Komoran
        import os
        if userdic is not None:
            print("user dict " + str(os.path.abspath(userdic)))
            self.inst = Komoran(userdic=os.path.abspath(userdic))
        else:
            self.inst = Komoran()
        self.OUT_TYPE = [list, tuple]

    def __call__(self, *args, **kwargs):
        return self.inst.pos(args[0])

class TwitterKorean(BaseTokenizer):
    def __init__(self):
        from konlpy.tag import Twitter
        self.inst = Twitter()

        self.OUT_TYPE = [list, tuple]

    def __call__(self, *args, **kwargs):
        return self.inst.pos(args[0])

class KokomaKorean(BaseTokenizer):
    def __init__(self):
        from konlpy.tag import Kkma
        self.inst = Kkma()

        self.OUT_TYPE = [list, tuple]

    def __call__(self, *args, **kwargs):
        return self.inst.pos(args[0])

class MeCab(BaseTokenizer):
    def __init__(self, path=None):
        #import MeCab
        #self.inst = MeCab.Tagger()

        from konlpy.tag import Mecab
        self.inst = Mecab(path)

        self.OUT_TYPE = [list, tuple]

    def __call__(self, *args, **kwargs):
        try:
            return self.inst.pos(args[0])
        except:
            return []

class SpecialTokenizer:
    IN_TYPE = [str]
    OUT_TYPE = [str]

class MaxScoreTokenizerKorean(SpecialTokenizer):
    def __init__(self, scores=None):
        from soynlp.tokenizer import MaxScoreTokenizer
        self.inst=MaxScoreTokenizer(scores=scores)
        self.OUT_TYPE = [list, str]

    def __call__(self, *args, **kwargs):
        tokens = self.inst.tokenize(args[0])
        return tokens

class LTokenizerKorean(SpecialTokenizer):
    def __init__(self, scores=None):
        from soynlp.tokenizer import LTokenizer
        self.inst=LTokenizer(scores=scores)

        self.OUT_TYPE = [list, str]

    def __call__(self, *args, **kwargs):
        tokens = self.inst.tokenize(args[0])
        return tokens

class RegexTokenizerKorean(SpecialTokenizer):
    def __init__(self):
        from soynlp.tokenizer import RegexTokenizer
        self.inst=RegexTokenizer()
        self.OUT_TYPE = [list, str]

    def __call__(self, *args, **kwargs):
        tokens=self.inst.tokenize(args[0])
        return tokens


import requests


class ColBasic:
    def __init__(self):
        self.userAgent = 'Mozilla/5.0 (X11; Linux x86_64; rv:10.0) Gecko/20100101 Firefox/10.0'
        self.sleepA = 9
        self.sleepB = 3

    def getInfo(params):
        return '기본 수집기', 0

    def getName(self):
        return "colBasic"

    def parseDate(self, date):
        import re
        m = re.match('([1-2][0-9]{3})([-.]?)([0-2][0-9])\\2([0-3][0-9])', date)
        if m: return (int(m.group(1)), int(m.group(3)), int(m.group(4)))
        return None

    def reformatDate(self, date, format):
        d = self.parseDate(date)
        if not d: return ''
        return format % d

    def initParameter(self, params):
        startDate = params['startDate']
        endDate = params['endDate']
        if params.get('lastDate'): endDate = params['lastDate']
        startDate = self.reformatDate(startDate, '%04d.%02d.%02d')
        endDate = self.reformatDate(endDate, '%04d.%02d.%02d')
        return {'startDate':startDate, 'endDate':endDate}

    def fetchList(self, page, query, d):
        import urllib.parse, urllib.request, bs4
        data = urllib.parse.urlencode(self.makeParameter(page, query, d))
        #req = urllib.request.Request(self.getSearchURL() + '?' + data)
        header = {'Referer' : self.getSearchURL(),
                   'User-Agent' : self.userAgent}

        page = requests.get(self.getSearchURL() + '?' + data, headers=header)
        page.encoding = 'utf-8'
        if page.status_code == 200:
            soup = bs4.BeautifulSoup(page.content, "lxml")

        return self.selectList(soup)

    @staticmethod
    def cleanText(t):
        return t.replace('\n', ' ').replace('\t', ' ').replace('\r', ' ')

    def fetchArticle(self, id):
        import urllib.request, bs4, traceback, time, random
        try:
            req = urllib.request.Request(id)
            req.add_header('Referer', self.getSearchURL())
            req.add_header('User-Agent', self.userAgent)
            f = urllib.request.urlopen(req)
            bytestream = f.read()
            try:
                cont = bytestream.decode('utf-8')
            except:
                cont = bytestream.decode('euc-kr')
            soup = bs4.BeautifulSoup(cont, "lxml")
            return (id, ) + self.selectArticle(soup)
        except:
            traceback.print_exc()
            time.sleep(random.random() * self.sleepA + self.sleepB + 1)

    def collect(self, query, outputPath, **params):
        import sys, traceback
        import time, random
        try:
            page = int(params['page'])
        except:
            page = 1
        d = self.initParameter(params)
        output = open(outputPath, 'w', encoding='utf-8')
        print(d)
        try:
            while 1:
                urls = self.fetchList(page, query, d)
                print(urls[0])
                fetched = 0
                for url in urls:
                    article = self.fetchArticle(url)
                    if not article: continue
                    fetched += 1
                    print(article)
                    output.write('\t'.join(map(str, article)) + '\n')
                    output.flush()
                    time.sleep(random.random() * self.sleepA + self.sleepB)
                if fetched == 0 and len(urls) > 5:
                    raise RuntimeError('Failed to fetch articles \n' + '\n'.join(urls))
                page += 1
        except:
            traceback.print_exc()
        output.close()

    def getSearchURL(self):
        pass

    def makeParameter(self, page, query, d):
        pass

    def selectList(self, soup):
        pass

    def selectArticle(self, soup):
        pass

    @staticmethod
    def getTextFromElement(soup):
        if not soup: return None
        [s.extract() for s in soup('script')]
        [s.extract() for s in soup('style')]
        return ColBasic.cleanText(soup.text)


import gzip
import math
import os
import re
import sys
import time
import traceback
import logging

import numpy as np
#import tensorflow as tf
import tensorflow.compat.v1 as tf

from treform.utility import *

from treform.utility.base_util import is_server, is_my_pc, is_my_gpu_pc
from treform.utility.datafile_util import DataFileUtil
from treform.utility.dataset import DataSet
from treform.utility.hangul_util import HangulUtil
from treform.utility.num_util import NumUtil
from treform.utility.char_one_hot_vector import CharOneHotVector
from treform.utility.string_util import StringUtil
from treform.utility.watch_util import WatchUtil
from treform.utility.log_util import LogUtil


from treform.spelling.config import KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR, \
    KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE, KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE, KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE, KO_WIKIPEDIA_ORG_CHARACTERS_FILE, \
    KO_WIKIPEDIA_ORG_DIR

log = LogUtil.get_logger(source_filepath=None, level=logging.INFO, console_mode=True)

tf.disable_eager_execution()

class SpellingErrorCorrection(object):
    graph = {}

    @classmethod
    def learning(cls, total_epoch, n_train, n_valid, n_test, batch_size, window_size, noise_rate, noise_sampling, model_file, features_vector, labels_vector,
                 n_hidden1,
                 learning_rate,
                 dropout_keep_rate):
        n_features = len(features_vector) * window_size  # number of features = 17,382 * 10

        log.info('load characters list...')
        log.info('load characters list OK. len: %s' % NumUtil.comma_str(len(features_vector)))
        watch = WatchUtil()

        train_file = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'datasets', 'spelling_error_correction',
                                  'ko.wikipedia.org.dataset.sentences=%s.window_size=%d.train.gz' % (n_train, window_size))
        valid_file = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'datasets', 'spelling_error_correction',
                                  'ko.wikipedia.org.dataset.sentences=%s.window_size=%d.train.gz' % (n_valid, window_size))
        test_file = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'datasets', 'spelling_error_correction',
                                 'ko.wikipedia.org.dataset.sentences=%s.window_size=%d.train.gz' % (n_test, window_size))
        if is_my_pc() or is_my_gpu_pc() or not os.path.exists(train_file) or not os.path.exists(valid_file) or not os.path.exists(test_file):
            dataset_dir = os.path.dirname(train_file)
            if not os.path.exists(dataset_dir):
                os.makedirs(dataset_dir)

            watch.start('create dataset')
            log.info('create dataset...')

            if is_server():
                data_files = (('train', KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE, n_train, train_file, False),
                              ('valid', KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE, n_valid, valid_file, False),
                              ('test', KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE, n_test, test_file, False))
            else:
                data_files = (('train', KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE, n_train, train_file, False),)

            for (name, data_file, total, dataset_file, to_one_hot_vector) in (data_files):
                check_interval = max(1, min(10000, batch_size // 10))
                log.info('check_interval: %s' % check_interval)
                log.info('%s %s total: %s' % (name, os.path.basename(data_file), NumUtil.comma_str(total)))
                log.info('noise_rate: %s' % noise_rate)

                features, labels = [], []
                with gzip.open(data_file, 'rt', encoding='utf-8') as f:
                    for i, line in enumerate(f, 1):
                        if total < i:
                            break

                        if i % check_interval == 0:
                            time.sleep(0.01)  # prevent cpu overload
                            percent = i / total * 100
                            log.info('create dataset... %.1f%% readed. data len: %s. %s' % (percent, NumUtil.comma_str(len(features)), data_file))

                        sentence = line.strip()
                        for start in range(0, len(sentence) - window_size + 1):  # 문자 단위로 노이즈(공백) 생성
                            chars = sentence[start: start + window_size]
                            for idx in range(len(chars)):
                                noised_chars = StringUtil.replace_with_index(chars, ' ', idx)
                                features.append(noised_chars)
                                labels.append(chars)
                                log.debug('create dataset... %s "%s" -> "%s"' % (name, noised_chars, chars))

                # log.info('noise_sampling: %s' % noise_sampling)
                #         for nth_sample in range(noise_sampling): # 초성, 중성, 종성 단위로 노이즈 생성
                #             for start in range(0, len(sentence) - window_size + 1):
                #                 chars = sentence[start: start + window_size]
                #                 noised_chars = SpellingErrorCorrection.encode_noise(chars, noise_rate=noise_rate, noise_with_blank=True)
                #                 if chars == noised_chars:
                #                     continue
                #                 if i % check_interval == 0 and nth_sample == 0:
                #                     log.info('create dataset... %s "%s" -> "%s"' % (name, noised_chars, chars))
                #                 features.append(noised_chars)
                #                 labels.append(chars)

                # print('dataset features:', features)
                # print('dataset labels:', labels)
                dataset = DataSet(features=features, labels=labels, features_vector=features_vector, labels_vector=labels_vector, name=name)
                log.info('dataset save... %s' % dataset_file)
                dataset.save(dataset_file, gzip_format=True, verbose=True)
                log.info('dataset save OK. %s' % dataset_file)
                log.info('dataset: %s' % dataset)

            log.info('create dataset OK.')
            log.info('')
            watch.stop('create dataset')

        watch.start('dataset load')
        log.info('dataset load...')
        train = DataSet.load(train_file, gzip_format=True, verbose=True)

        if is_server():
            valid = DataSet.load(valid_file, gzip_format=True, verbose=True)
        else:
            valid = DataSet.load(train_file, gzip_format=True, verbose=True)  # valid with train set
        log.info('valid.convert_to_one_hot_vector()...')
        valid = valid.convert_to_one_hot_vector(verbose=True)
        log.info('valid.convert_to_one_hot_vector() OK.')

        log.info('train dataset: %s' % train)
        log.info('valid dataset: %s' % valid)
        log.info('dataset load OK.')
        log.info('')
        watch.stop('dataset load')

        X, Y, dropout_keep_prob, train_step, cost, y_hat, accuracy = SpellingErrorCorrection.build_DAE(n_features, window_size, noise_rate, n_hidden1,
                                                                                                       learning_rate, watch)

        with tf.Session() as sess:
            sess.run(tf.global_variables_initializer())
            check_interval = max(1, min(1000, n_train // 10))

            nth_train, nth_input, total_input = 0, 0, total_epoch * train.size
            log.info('')
            log.info('learn...')
            log.info('total_epoch: %s' % total_epoch)
            log.info('train.size (total features): %s' % NumUtil.comma_str(train.size))
            log.info('check_interval: %s' % check_interval)
            log.info('total_epoch: %s' % total_epoch)
            log.info('batch_size: %s' % batch_size)
            log.info('total_input: %s (total_epoch * train.size)' % total_input)
            log.info('')
            watch.start('learn')
            for epoch in range(1, total_epoch + 1):
                for step, (features_batch, labels_batch) in enumerate(train.next_batch(batch_size=batch_size, to_one_hot_vector=True), 1):
                    nth_train += 1
                    nth_input += features_batch.shape[0]
                    sess.run(train_step, feed_dict={X: features_batch, Y: labels_batch, dropout_keep_prob: dropout_keep_rate})

                    percent = nth_input / total_input * 100

                    # if nth_train % check_interval == 1:
                    valid_cost = sess.run(cost, feed_dict={X: valid.features, Y: valid.labels, dropout_keep_prob: 1.0})
                    log.info('[epoch=%s][%.1f%%] %s cost: %.8f' % (epoch, percent, valid.name, valid_cost))

            watch.stop('learn')
            log.info('learn OK.')
            log.info('')

            log.info('model save... %s' % model_file)
            watch.start('model save...')
            model_dir = os.path.dirname(model_file)
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
            saver = tf.train.Saver()
            saver.save(sess, model_file)
            watch.stop('model save...')
            log.info('model save OK. %s' % model_file)

        log.info('')
        log.info('total_epoch: %s' % total_epoch)
        log.info('batch_size: %s' % batch_size)
        log.info('total_input: %s (total_epoch * train.size)' % total_input)
        log.info('')
        log.info(watch.summary())
        log.info('')

    @classmethod
    def build_DAE(cls, n_features, window_size, noise_rate, n_hidden1, learning_rate, watch=WatchUtil()):
        if len(cls.graph) == 0:
            log.info('')
            log.info('create tensorflow graph...')
            watch.start('create tensorflow graph')

            features_vector_size = n_features // window_size
            log.info('n_features: %s' % n_features)
            log.info('window_size: %s' % window_size)
            log.info('features_vector_size: %s' % features_vector_size)

            log.info('noise_rate: %.1f' % noise_rate)
            log.info('n_hidden1: %s' % n_hidden1)

            tf.set_random_seed(777)  # for reproducibility

            X = tf.placeholder(tf.float32, [None, n_features], name='X')  # shape=(batch_size, window_size * feature_vector.size)
            Y = tf.placeholder(tf.float32, [None, n_features], name='Y')  # shape=(batch_size, window_size * feature_vector.size)
            dropout_keep_prob = tf.placeholder(tf.float32)

            # layers = 3
            # n_hidden2 = n_hidden1
            # W1 = tf.Variable(tf.random_normal([n_features, n_hidden1]), name='W1')
            # b1 = tf.Variable(tf.random_normal([n_hidden1]), name='b1')
            # layer1 = tf.nn.sigmoid(tf.matmul(X, W1) + b1, name='layer1')
            # layer1_dropout = tf.nn.dropout(layer1, dropout_keep_prob, name='layer1_dropout')
            #
            # W2 = tf.Variable(tf.random_normal([n_hidden1, n_hidden2]), name='W2')
            # b2 = tf.Variable(tf.random_normal([n_hidden2]), name='b2')
            # layer2 = tf.nn.sigmoid(tf.matmul(layer1_dropout, W2) + b2, name='layer2')
            # layer2_dropout = tf.nn.dropout(layer2, dropout_keep_prob, name='layer2_dropout')
            #
            # W3 = tf.Variable(tf.random_normal([n_hidden1, n_features]), name='W3')
            # b3 = tf.Variable(tf.random_normal([n_features]), name='b3')
            # y_hat = tf.add(tf.matmul(layer2_dropout, W3), b3, name='y_hat')

            # layers = 2
            W1 = tf.Variable(tf.random_normal([n_features, n_hidden1]), name='W1')
            b1 = tf.Variable(tf.random_normal([n_hidden1]), name='b1')
            layer1 = tf.nn.sigmoid(tf.matmul(X, W1) + b1, name='layer1')
            layer1_dropout = tf.nn.dropout(layer1, dropout_keep_prob, name='layer1_dropout')

            W2 = tf.Variable(tf.random_normal([n_hidden1, n_features]), name='W2')
            b2 = tf.Variable(tf.random_normal([n_features]), name='b2')
            y_hat = tf.add(tf.matmul(layer1_dropout, W2), b2, name='y_hat')  # shape=(batch_size, window_size * feature_vector.size)

            labels_hat = tf.reshape(y_hat, shape=(-1, window_size, features_vector_size))  # shape=(batch_size, window_size, feature_vector.size)
            labels = tf.reshape(Y, shape=(-1, window_size, features_vector_size))  # shape=(batch_size, window_size, feature_vector.size)

            cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=labels_hat, labels=labels), name='cost')
            train_step = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cost)

            accuracy = tf.reduce_mean(tf.cast(tf.abs(tf.nn.softmax(y_hat) - Y) < 0.1, dtype=tf.float32), name='accuracy')
            # log.debug('X:', X)
            # log.debug('Y:', Y)
            # log.debug('y_hat:', y_hat)
            # log.debug('labels_hat:', labels_hat)
            # log.debug('labels:', labels)
            # log.debug('cost:', cost)
            # log.debug('accuracy:', accuracy)

            watch.stop('create tensorflow graph')
            log.info('create tensorflow graph OK.')
            log.info('')
            cls.graph = {'X': X, 'Y': Y, 'dropout_keep_prob': dropout_keep_prob,
                         'train_step': train_step, 'cost': cost, 'y_hat': y_hat, 'accuracy': accuracy, }
        return cls.graph['X'], cls.graph['Y'], cls.graph['dropout_keep_prob'], \
               cls.graph['train_step'], cls.graph['cost'], cls.graph['y_hat'], cls.graph['accuracy']

    @classmethod
    def sentence2features_labels(cls, sentence, noised_sentence, window_size=10, skip_same=False) -> (list, list):
        try:
            sentence, noised_sentence = sentence.strip(), noised_sentence.strip()
            if window_size < 1 or len(sentence) != len(noised_sentence):  # or len(sentence) <= 1 :
                return [], []

            features, labels = [], []
            for start in range(0, len(sentence) - window_size + 1):
                feature, label = noised_sentence[start:start + window_size], sentence[start:start + window_size]
                if skip_same and feature != label:
                    continue

                features.append(feature)
                labels.append(label)
            return features, labels
        except:
            return [], []

    @classmethod
    def encode_noise(cls, s, noise_rate=0.1, noise_with_blank=False, verbose=False):
        try:
            hangul_indexs = [idx for idx, c in enumerate(s) if HangulUtil.is_hangul_char(c)]
            if len(hangul_indexs) == 0:
                return s
            target_indexs = np.random.choice(hangul_indexs, math.ceil(len(hangul_indexs) * noise_rate), replace=False)
            _s = list(s)
            for idx in target_indexs:
                if noise_with_blank:
                    _s[idx] = ' '
                else:
                    c = s[idx]
                    _c = HangulUtil.encode_noise(c)
                    if verbose:
                        log.info('encode: %s -> %s' % (c, _c))
                    _s[idx] = _c
            return ''.join(_s)
        except:
            return s

    @classmethod
    def decode_noise(cls, noised_sentence, features_list, labels_list, verbose=False):
        try:
            if len(features_list) != len(labels_list) or len(features_list[0]) != len(labels_list[0]):
                return noised_sentence

            idx2chars = dict()
            for feature, label in zip(features_list, labels_list):
                for off in [i for i in range(len(feature)) if feature[i] != label[i]]:
                    for start in [m.start() for m in re.finditer(feature, noised_sentence)]:
                        idx = start + off
                        if idx not in idx2chars:
                            idx2chars[idx] = label[off]
            sentence = list(noised_sentence)
            for idx, char in idx2chars.items():
                if verbose:
                    log.info('denoise: "%s" -> "%s"' % (noised_sentence[idx], char))
                sentence[idx] = char
            return ''.join(sentence)
        except:
            return noised_sentence

    @classmethod
    def sim_two_sentence(cls, sentence, sentence_hat):
        sim, correct, total = 0, 0, len(sentence)
        if len(sentence) != len(sentence_hat):
            return sim, correct, total

        for a, b in zip(sentence, sentence_hat):
            if a == b:
                correct += 1

        sim = correct / total
        return sim, correct, total


if __name__ == '__main__':
    train_sentences_file = KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE
    valid_sentences_file = KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE
    test_sentences_file = KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE
    log.info('train_sentences_file: %s' % train_sentences_file)
    log.info('valid_sentences_file: %s' % valid_sentences_file)
    log.info('test_sentences_file: %s' % test_sentences_file)
    log.info('')

    characters_file = KO_WIKIPEDIA_ORG_CHARACTERS_FILE
    log.info('characters_file: %s' % characters_file)
    try:
        if len(sys.argv) == 4:
            n_train = int(sys.argv[1])
            window_size = int(sys.argv[2])
            noise_rate = float(sys.argv[3])
        else:
            n_train, noise_rate, window_size = None, None, None

        if n_train is None or n_train == 0:
            n_train = int('1,000,000'.replace(',', ''))

        if is_server():  # batch server
            n_train = 10000
            n_valid = min(10, n_train)
            n_test = min(10, n_train)
        else:  # for demo
            #n_train = 10000
            #n_valid = min(10, n_train)
            #n_test = min(10, n_train)
            n_train = n_valid = n_test = 3

        if noise_rate is None or window_size is None:
            window_size = 10  # 2 ~ 10 # feature로 추출할 문자 수 (label과 동일)
            noise_rate = max(0.1, 1 / window_size)  # 0.0 ~ 1.0 # noise_rate = 노이즈 문자 수 / 전체 문자 수 (windos 안에서 최소 한 글자는 노이즈가 생기도록 함.)

        dropout_keep_rate = 1.0  # 0.0 ~ 1.0 # one hot vector에 경우에 dropout 사용시, 학습이 안 됨.
        noise_sampling = 100  # 한 입력에 대하여 몇 개의 노이즈 샘플을 생성할지. blank 방식(문자 단위)으로 noise 생성할 때는 사용 안함.

        total_epoch = max(10, 1000 // window_size)  # 10 ~ 100 # window_size 가 클 수록 total_epoch는 작아도 됨.
        batch_size = min(100, 10 * window_size)  # 1 ~ 100 # one hot vector 입력이면, batch_size 작게 잡아야 학습이 잘 된다. batch_size가 너무 크면, 전혀 학습이 안 됨.

        n_hidden1 = min(1000, 10 * window_size)  # 10 ~ 1000
        learning_rate = 1 / total_epoch  # 0.01  # 0.1 ~ 0.001  # total_epoch 가 클 수록 learning_rate는 작아도 됨.

        log.info('')
        log.info('n_train (sentences): %s' % NumUtil.comma_str(n_train))
        log.info('n_valid (sentences): %s' % NumUtil.comma_str(n_valid))
        log.info('n_test (sentences): %s' % NumUtil.comma_str(n_test))
        log.info('')
        log.info('window_size: %s' % window_size)
        log.info('noise_rate: %s' % noise_rate)
        log.info('dropout_keep_rate: %s' % dropout_keep_rate)
        log.info('noise_sampling: %s' % noise_sampling)
        log.info('')
        log.info('n_hidden1: %s' % n_hidden1)
        log.info('learning_rate: %s' % learning_rate)
        log.info('')
        log.info('total_epoch: %s' % total_epoch)
        log.info('batch_size: %s' % batch_size)

        model_file = os.path.join(KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR,
                                  'spelling_error_correction_model.sentences=%s.window_size=%s.noise_rate=%.1f.n_hidden=%s/model' % (
                                      n_train, window_size, noise_rate, n_hidden1))  # .%s' % max_sentences
        log.info('model_file: %s' % model_file)

        features_vector = CharOneHotVector(DataFileUtil.read_list(characters_file))
        labels_vector = CharOneHotVector(DataFileUtil.read_list(characters_file))
        n_features = len(features_vector) * window_size  # number of features = 17,450 * 10
        n_classes = len(labels_vector) * window_size  # number of features = 17,450 * 10
        log.info('')
        log.info('features_vector: %s' % features_vector)
        log.info('labels_vector: %s' % labels_vector)
        log.info('n_features (input vector size): %s' % n_features)
        log.info('n_classes (input vector size): %s' % n_classes)

        log.info('')
        log.info('sample testing...')
        DEMO_SENTENCES = ['아버지가 방에 들어 가신다.', '오빠가 방에 들어 간다.', '이모가 방에 들어 가신다.']
        for i, sentence in enumerate(DEMO_SENTENCES):
            noised_sentence = SpellingErrorCorrection.encode_noise(sentence, noise_rate=0.1, noise_with_blank=True)
            log.info('[%s] ori: "%s"' % (i, sentence))
            _features, _labels = SpellingErrorCorrection.sentence2features_labels(sentence, noised_sentence, window_size)
            for _f, _l in zip(_features, _labels):
                log.info('learn "%s" -> "%s"' % (_f, _l))

            sentence_hat = SpellingErrorCorrection.decode_noise(noised_sentence, _features, _labels)
            log.info('[%s] in : "%s"' % (i, noised_sentence))

            sim, correct, total = SpellingErrorCorrection.sim_two_sentence(sentence, sentence_hat)
            log.info('[%s] out: "%s" (sim: %.1f%%=%s/%s)' % (i, sentence_hat, sim * 100, correct, total))
        log.info('sample testing OK.')
        log.info('')

        # if is_my_pc() or is_my_gpu_pc() or not os.path.exists(model_file + '.index') or not os.path.exists(model_file + '.meta'):
        if not os.path.exists(model_file + '.index') or not os.path.exists(model_file + '.meta'):
            if n_train > int('100,000'.replace(',', '')):
                #SlackUtil.send_message('%s start (max_sentences=%s, window_size=%s, noise_rate=%.1f)' % (sys.argv[0], n_train, window_size, noise_rate))
                print('%s start (max_sentences=%s, window_size=%s, noise_rate=%.1f)' % (sys.argv[0], n_train, window_size, noise_rate))
            SpellingErrorCorrection.learning(total_epoch, n_train, n_valid, n_test, batch_size, window_size, noise_rate, noise_sampling,
                                             model_file, features_vector, labels_vector, n_hidden1=n_hidden1,
                                             learning_rate=learning_rate, dropout_keep_rate=dropout_keep_rate)
            if n_train > int('100,000'.replace(',', '')):
                #SlackUtil.send_message('%s end (max_sentences=%s, window_size=%s, noise_rate=%.1f)' % (sys.argv[0], n_train, window_size, noise_rate))
                print('%s end (max_sentences=%s, window_size=%s, noise_rate=%.1f)' % (sys.argv[0], n_train, window_size, noise_rate))
        log.info('chek result...')
        watch = WatchUtil()
        watch.start('read sentences')

        max_test_sentences = 1
        sentences = []
        if is_server():
            sentences_file = test_sentences_file
        else:
            sentences_file = train_sentences_file

        with gzip.open(sentences_file, 'rt', encoding='utf-8') as f:
            for i, line in enumerate(f, 1):
                if len(sentences) >= n_test:
                    break

                s = line.strip()
                sentences.append(s)
        log.info('len(sentences): %s' % NumUtil.comma_str(len(sentences)))
        watch.stop('read sentences')

        watch.start('run tensorflow')

        with tf.Session() as sess:
            X, Y, dropout_keep_prob, train_step, cost, y_hat, accuracy = SpellingErrorCorrection.build_DAE(n_features, window_size, noise_rate, n_hidden1,
                                                                                                           learning_rate, watch)
            saver = tf.train.Saver()
            try:
                restored = saver.restore(sess, model_file)
            except Exception as e:
                log.error('restore failed. model_file: %s' % model_file)
                raise e

            train_file = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'datasets', 'spelling_error_correction',
                                      'ko.wikipedia.org.dataset.sentences=%s.window_size=%d.train.gz' % (n_train, window_size))
            train = DataSet.load(train_file, gzip_format=True, verbose=True)
            train_vector = DataSet.load(train_file, gzip_format=True, verbose=True)
            train_vector.convert_to_one_hot_vector()
            try:
                accuracies, costs, sims = [], [], []
                total_test_sampling = 1
                for i, sentence in enumerate(sentences):
                    for nth in range(total_test_sampling):
                        # log.info('[%s] noise(%.1f) "%s" -> "%s"' % (nth, noise_rate, sentence, noised_sentence))
                        noised_sentence = SpellingErrorCorrection.encode_noise(sentence, noise_rate=0.1)
                        log.info('')
                        log.info('[%s] ori: "%s"' % (nth, sentence))
                        log.info('[%s] in : "%s"' % (nth, noised_sentence))

                        denoised_sentence = noised_sentence[:]  # will be changed with predict
                        for start in range(0, len(noised_sentence) - window_size + 1):
                            chars = denoised_sentence[start: start + window_size]
                            original_chars = sentence[start: start + window_size]
                            _features = [chars]
                            _labels = [original_chars]

                            dataset = DataSet(features=_features, labels=_labels, features_vector=features_vector, labels_vector=features_vector)
                            dataset.convert_to_one_hot_vector()
                            try:
                                _y_hat, _cost, _accuracy = sess.run([y_hat, cost, accuracy],
                                                         feed_dict={X: dataset.features, Y: dataset.labels, dropout_keep_prob: dropout_keep_rate})
                                costs.append(_cost)
                                accuracies.append(_accuracy)
                            except:
                                log.error('"%s"%s "%s"%s' % (chars, dataset.features.shape, original_chars, dataset.labels.shape))

                            y_hats = [features_vector.to_values(_l) for _l in _y_hat]
                            if _features[0] == y_hats[0]:
                                log.info('same   : "%s"' % (_features[0]))
                            else:
                                log.info('denoise: "%s" -> "%s"' % (_features[0], y_hats[0]))
                            denoised_sentence = denoised_sentence.replace(_features[0], y_hats[0])
                            # print(denoised_sentence)

                        sim, correct, total = SpellingErrorCorrection.sim_two_sentence(sentence, denoised_sentence)
                        sims.append(sim)

                        if correct == total:
                            log.info('[%s] out: "%s" (cost: %.4f, accuracy: %.4f, sim: %.1f%%=%s/%s)' % (
                                nth, denoised_sentence, _cost, _accuracy, sim * 100, correct, total))
                        else:
                            log.error('[%s] out: "%s" (cost: %.4f, accuracy: %.4f, sim: %.1f%%=%s/%s) incorrect!!' % (
                                nth, denoised_sentence, _cost, _accuracy, sim * 100, correct, total))
            except Exception as e:
                raise e

        #tf.contrib.simple_save(sess, export_dir, inputs, outputs)

        log.info('chek result OK.')
        # noinspection PyStringFormat
        log.info('mean(cost): %.4f, mean(sim): %.2f%%' % (np.mean(costs), np.mean(sims) * 100))
        log.info('secs/sentence: %.4f' % (watch.elapsed('run tensorflow') / len(sentences)))
        log.info('total_epoch: %s, batch_size: %s' % (NumUtil.comma_str(total_epoch), NumUtil.comma_str(batch_size)))
        log.info(watch.summary())
    except:
        log.error(traceback.format_exc())


import json

from pycrfsuite_spacing import PyCRFSuiteSpacing
from pycrfsuite_spacing import TemplateGenerator
from pycrfsuite_spacing import CharacterFeatureTransformer

from treform.segmentation.cnnWordSegmentationModel import SpacingModel

import tensorflow as tf


class BaseSegmentation:
    IN_TYPE = [str]
    OUT_TYPE = [str]


class SegmentationKorean(BaseSegmentation):
    def __init__(self, model=None):
        # model_path = 'demo_model.crfsuite'
        to_feature = CharacterFeatureTransformer(
            TemplateGenerator(
                begin=-2,
                end=2,
                min_range_length=3,
                max_range_length=3)
        )
        self.inst = PyCRFSuiteSpacing(to_feature)
        self.inst.load_tagger(model)

    def __call__(self, *args, **kwargs):
        return self.inst(args[0])


from treform.segmentation.lstmWordSegmentationModel import *
from treform.segmentation.wordSegmentationModelUtil import *


class LSTMSegmentationKorean(BaseSegmentation):
    def __init__(self, model_path='./model'):
        self.model = model_path
        dic_path = self.model + '/' + 'dic.pickle'

        # config
        self.n_steps = 30  # time steps
        self.padd = '\t'  # special padding chracter
        with open(dic_path, 'rb') as handle:
            self.char_dic = pickle.load(handle)  # load dic
        n_input = len(self.char_dic)  # input dimension, vocab size
        self.n_hidden = 8  # hidden layer size
        self.n_classes = 2  # output classes,  space or not
        self.vocab_size = n_input

        self.x = tf.placeholder(tf.float32, [None, self.n_steps, n_input])
        self.y_ = tf.placeholder(tf.int32, [None, self.n_steps])
        self.early_stop = tf.placeholder(tf.int32)

        # LSTM layer
        # 2 x n_hidden = state_size = (hidden state + cell state)
        self.istate = tf.placeholder(tf.float32, [None, 2 * self.n_hidden])
        weights = {
            'hidden': weight_variable([n_input, self.n_hidden]),
            'out': weight_variable([self.n_hidden, self.n_classes])
        }
        biases = {
            'hidden': bias_variable([self.n_hidden]),
            'out': bias_variable([self.n_classes])
        }

        self.y = RNN(self.x, self.istate, weights, biases, self.n_hidden, self.n_steps, n_input, self.early_stop)

        self.batch_size = 1
        self.logits = tf.reshape(tf.concat(self.y, 1), [-1, self.n_classes])

        NUM_THREADS = 1
        config = tf.ConfigProto(intra_op_parallelism_threads=NUM_THREADS,
                                inter_op_parallelism_threads=NUM_THREADS,
                                log_device_placement=False)
        self.sess = tf.Session(config=config)
        init = tf.global_variables_initializer()
        self.sess.run(init)
        saver = tf.train.Saver()  # save all variables
        checkpoint_dir = self.model
        ckpt = tf.train.get_checkpoint_state(checkpoint_dir)
        if ckpt and ckpt.model_checkpoint_path:
            saver.restore(self.sess, ckpt.model_checkpoint_path)
            sys.stderr.write("model restored from %s\n" % (ckpt.model_checkpoint_path))
        else:
            sys.stderr.write("no checkpoint found" + '\n')
            sys.exit(-1)

    def __call__(self, *args, **kwargs):

        sentence = args[0]
        sentence_size = len(sentence)
        tag_vector = [-1] * (sentence_size + self.n_steps)  # buffer n_steps
        pos = 0
        while pos != -1:
            batch_xs, batch_ys, next_pos, count = next_batch(sentence, pos, self.char_dic, self.vocab_size,
                                                             self.n_steps, self.padd)
            '''
            print 'window : ' + sentence[pos:pos+n_steps]
            print 'count : ' + str(count)
            print 'next_pos : ' + str(next_pos)
            print batch_ys
            '''
            c_istate = np.zeros((self.batch_size, 2 * self.n_hidden))
            feed = {self.x: batch_xs, self.y_: batch_ys, self.istate: c_istate, self.early_stop: count}
            argmax = tf.arg_max(self.logits, 1)
            result = self.sess.run(argmax, feed_dict=feed)

            # overlapped copy and merge
            j = 0
            result_size = len(result)
            while j < result_size:
                tag = result[j]
                if tag_vector[pos + j] == -1:
                    tag_vector[pos + j] = tag
                else:
                    if tag_vector[pos + j] == CLASS_1:  # 1
                        if tag == CLASS_0:  # 1 -> 0
                            sys.stderr.write("1->0\n")
                            tag_vector[pos + j] = tag
                    else:  # 0
                        if tag == CLASS_1:  # 0 -> 1
                            sys.stderr.write("0->1\n")
                            tag_vector[pos + j] = tag
                j += 1
            pos = next_pos
        # generate output using tag_vector
        print
        'out = ' + to_sentence(tag_vector, sentence)

        return to_sentence(tag_vector, sentence)

    def close(self):
        self.sess.close()


class CNNSegmentationKorean(BaseSegmentation):
    def __init__(self, model_file='./model', training_config='', char_file=''):
        self.model_file = model_file
        self.training_config = training_config
        self.char_file = char_file

    def load(self, model_file='', training_config='', char_file=''):
        tf.config.run_functions_eagerly(True)

        print(tf.executing_eagerly())

        with open(self.training_config, encoding='utf-8') as f:
            config = json.load(f)

        with open(self.char_file, encoding='utf=8') as f:
            content = f.read()
            keys = ["<pad>", "<s>", "</s>", "<unk>"] + list(content)
            values = list(range(len(keys)))

            vocab_initializer = tf.lookup.KeyValueTensorInitializer(keys, values, key_dtype=tf.string,
                                                                    value_dtype=tf.int32)
            vocab_table = tf.lookup.StaticHashTable(vocab_initializer, default_value=3)

        model = SpacingModel(
            config["vocab_size"],
            config["hidden_size"],
            conv_activation=config["conv_activation"],
            dense_activation=config["dense_activation"],
            conv_kernel_and_filter_sizes=config["conv_kernel_and_filter_sizes"],
            dropout_rate=config["dropout_rate"],
        )

        model.load_weights(self.model_file)
        model(tf.keras.Input([None], dtype=tf.int32))
        model.summary()

        return model, vocab_table

    def predict(self, model, vocab_table, input_str):
        inference = self.get_inference_fn(model, vocab_table)
        input_str = tf.constant(input_str)
        result1 = inference(input_str)
        sess = tf.compat.v1.Session()
        init = tf.compat.v1.global_variables_initializer()
        sess.run(init)
        sess.run(tf.compat.v1.tables_initializer())

        result = sess.run(inference(input_str))
        print(result)

        print(b"".join(result).decode("utf8"))
        return b"".join(result).decode("utf8")

    def get_inference_fn(self, model, vocab_table):
        @tf.function
        def inference(tensors):
            byte_array = tf.concat(
                [["<s>"], tf.strings.unicode_split(tf.strings.regex_replace(tensors, " +", " "), "UTF-8"), ["</s>"]],
                axis=0
            )
            strings = vocab_table.lookup(byte_array)[tf.newaxis, :]

            model_output = tf.argmax(model(strings), axis=-1)[0]
            return self.convert_output_to_string(byte_array, model_output)

        return inference

    def convert_output_to_string(self, byte_array, model_output):
        sequence_length = tf.size(model_output)
        while_condition = lambda i, *_: i < sequence_length

        #@tf.function
        def while_body(i, o):
            a = tf.math.equal(model_output[i],2)
            b = tf.math.equal(byte_array[i],' ')
            c = tf.math.equal(model_output[i],1)

            o = tf.cond(
                tf.cast(c, tf.bool),
                lambda: tf.concat([o, [byte_array[i], " "]], axis=0),
                lambda: tf.cond(
                    tf.cast(tf.logical_and(a, b),tf.bool),
                    lambda: o,
                    lambda: tf.concat([o, [byte_array[i]]], axis=0),
                ),
            )
            return i + 1, o

        _, strings_result = tf.while_loop(
            while_condition,
            while_body,
            (tf.constant(0), tf.constant([], dtype=tf.string)),
            shape_invariants=(tf.TensorShape([]), tf.TensorShape([None])),
        )
        return strings_result

    def __call__(self, *args, **kwargs):
        model, vocab_table = self.load(self.model_file, self.training_config, self.char_file)
        return self.predict(model, vocab_table, args[0])


from treform.collector.colBasic import *

class ColJoongang(ColBasic):
    def getInfo(params):
        return '중앙일보 수집기', 0

    def getName(self):
        return "colJoongang"

    def getSearchURL(self):
        return 'http://search.joins.com/JoongangNews'

    def makeParameter(self, page, query, d):
        return {'StartSearchDate': d['startDate'],
                'EndSearchDate': d['endDate'],
                'Keyword': query,
                'SortType': 'New',
                'SearchCategoryType': 'JoongangNews',
                'PeriodType': 'DirectInput',
                'ScopeType': 'All',
                'ServiceCode': '',
                'SourceGroupType': '',
                'ReporterCode': '',
                'ImageType': 'All',
                'JplusType': 'All',
                'BlogType': 'All',
                'ImageSearchType': 'Image',
                'MatchKeyword': '',
                'IncludeKeyword': '',
                'ExcluedeKeyword': '',
                'page':page,
        }

    def selectList(self, soup):
        ret = []
        for el in soup.select('.headline a'):
            link = el['href']
            ret.append(link)
        return ret

    def selectArticle(self, soup):
        return (ColBasic.getTextFromElement(soup.select_one('h1')),
                ColBasic.cleanText(soup.find('meta', {'property': 'article:published_time'})['content']),
                ColBasic.getTextFromElement(soup.select_one('.article_body')))



from treform.document_classification.ml_textclassification import documentClassifier
import treform as ptm

if __name__ == '__main__':
    document_classifier = documentClassifier()
    mecab_path = 'C:\\mecab\\mecab-ko-dic'
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.MeCab(mecab_path),
                            ptm.helper.POSFilter('NN*'),
                            ptm.helper.SelectWordOnly(),
                            ptm.ngram.NGramTokenizer(2, 2),
                            #ptm.tokenizer.LTokenizerKorean(),
                            ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt')
                            )

    #mode is either train or predict
    mode = 'train'
    if mode is 'train':
        input_file ='./data/3_class_naver_news.csv'
        # 1. text processing and representation
        corpus = ptm.CorpusFromFieldDelimitedFileForClassification(input_file,
                                                                   delimiter=',',
                                                                   doc_index=4,
                                                                   class_index=1,
                                                                   title_index=3)
        corpus.docs
        tups = corpus.pair_map
        class_list = []
        for id in tups:
            #print(tups[id])
            class_list.append(tups[id])

        result = pipeline.processCorpus(corpus)
        print('==  ==')

        documents = []
        for doc in result:
            document = ''
            for sent in doc:
                document += " ".join(sent)
            documents.append(document)

        document_classifier.preprocess(documents,class_list)

        #model_name = 0  -- RandomForestClassifier
        #model_name = 1  -- LinearSVC
        #model_name = 2  -- MultinomialNB
        #model_name = 3  -- LogisticRegression
        #model_name = 4  -- K-NN
        #model_name = 5  -- SGDClassifier
        X_train, X_test, y_train, y_test, y_pred, indices_test, model = document_classifier.train(model_index=1)

        print('training is finished')

        document_classifier.evaluate(y_test,y_pred,indices_test,model)
        document_classifier.save(model, model_name='./model/svm_classifier.model')
        document_classifier.saveVectorizer(model_name='./model/vectorizer.model')

    elif mode is 'predict':
        model=document_classifier.load('./model/svm_classifier.model')
        vectorizer_model=document_classifier.loadVectorizer(model_name='./model/vectorizer.model')
        document_classifier.predict(model,vectorizer_model)

        #7. prediction
        input = "../data/navernews.txt"
        corpus = ptm.CorpusFromFieldDelimitedFile(input,3)

        result = pipeline.processCorpus(corpus)
        print('==  ==')

        documents = []
        for doc in result:
            document = ''
            for sent in doc:
                document += " ".join(sent)
            documents.append(document)

        document_classifier.predict_realtime(model,vectorizer_model, documents)




from krwordrank.word import KRWordRank
from joblib import load

class BaseKeywordExtraction:
    IN_TYPE = [str]
    OUT_TYPE = [str]

class TextRankExtractor(BaseKeywordExtraction):
    def __init__(self, pos_tagger_name=None, mecab_path=None,
                 lang='ko', max=10,
                 stopwords=[], combined_keywords=False):
        import treform.keyword.textrank as tr
        self.inst = tr.TextRank(pos_tagger_name=pos_tagger_name,mecab_path=mecab_path,lang=lang,stopwords=stopwords)
        self.max=max
        self.combined_keywords = combined_keywords
    def __call__(self, *args, **kwargs):
        self.OUT_TYPE = [list, str]
        import nltk.tokenize
        sents = nltk.tokenize.sent_tokenize(*args)
        for sent in sents:
            self.inst.build_keywords(sent)
        return self.inst.get_keywords(self.max,self.combined_keywords)


class MLKeyphraseExtractor(BaseKeywordExtraction):
    def __init__(self, language='ko', model_name='../../models/svm_keyphrase.model'):
        self.model = load(model_name)
        self.language=language

    def __call__(self, *args, **kwargs):

        from treform.keyword.ml_keyword_builder import get_features
        from treform.keyword.ml_keyword_extractor import extract_candidate_keywords, \
            extract_candidate_keywords_for_training

        self.OUT_TYPE = [list, str]
        feature_list = []

        candidates = extract_candidate_keywords_for_training(args[0], language=self.language)
        set_candidates = set(candidates)
        unique_list = list(set_candidates)
        if len(candidates) > 0:
            feature_list.extend([(get_features(args[0], key, candidates, 0, language=self.language)) for key in set_candidates])
        #for features in feature_list:
        #print(set_candidates)
        #print(feature_list)
        preds = self.model.classify_many(feature_list)
        #self.model.prob_classify_many(feature_list)
        labels = self.model.labels()
        #print(labels)
        print(preds)
        keyphrases = []
        for i, each_pred in enumerate(preds):
            if each_pred == 1:
                keyphrase = unique_list[i]
                keyphrases.append(keyphrase)
        return keyphrases

class TextRankSummarizer(BaseKeywordExtraction):
    def __init__(self,pos_tagger_name=None,mecab_path=None,max=3):
        import treform.keyword.textrank as tr
        self.inst=tr.TextRank(pos_tagger_name=pos_tagger_name,mecab_path=mecab_path)
        self.max=max

    def __call__(self, *args, **kwargs):
        return self.inst.summarize(args[0],self.max)

class KeywordExtractionKorean(BaseKeywordExtraction):
    def __init__(self, min_count=2, max_length=10,
                 beta=0.85, max_iter=10, verbose=True, num_words=20):
        self.min_count=min_count
        self.max_length=max_length
        self.beta=beta
        self.max_iter=max_iter
        self.verbose=verbose
        self.num_words=num_words

        self.inst=KRWordRank(min_count, max_length,self.verbose)

    def __call__(self, *args, **kwargs):
        _num_keywords=10
        #print(str(args[0]) + "\n")
        keywords, rank, graph = self.inst.extract(args[0], self.beta, self.max_iter, self.num_words)

        return keywords


from pickle import load
from numpy import array
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.utils.vis_utils import plot_model
from keras.models import Model
from keras.layers import Input
from keras.layers import Dense
from keras.layers import Flatten
from keras.layers import Dropout
from keras.layers import Embedding
from keras.layers.convolutional import Conv1D
from keras.layers.convolutional import MaxPooling1D
from keras.layers.merge import concatenate

class sentimentModel:
    def __init__(self):
        name = 'sentimentModel'

    # load a clean dataset
    def load_dataset(self, filename):
        return load(open(filename, 'rb'))

    # fit a tokenizer
    def create_tokenizer(self, lines):
        tokenizer = Tokenizer()
        #print("XX " + str(lines))
        tokenizer.fit_on_texts(lines)

        return tokenizer

    # calculate the maximum document length
    def max_length(self, lines):
        return max([len(s.split()) for s in lines])

    # encode a list of lines
    def encode_text(self, tokenizer, lines, length):
        # integer encode
        encoded = tokenizer.texts_to_sequences(lines)
        # pad encoded sequences
        padded = pad_sequences(encoded, maxlen=length, padding='post')
        return padded


    # define the model
    def define_model(self, length, vocab_size):
        # channel 1
        inputs1 = Input(shape=(length,))
        embedding1 = Embedding(vocab_size, 100)(inputs1)
        conv1 = Conv1D(filters=32, kernel_size=4, activation='relu')(embedding1)
        drop1 = Dropout(0.5)(conv1)
        pool1 = MaxPooling1D(pool_size=2)(drop1)
        flat1 = Flatten()(pool1)
        # channel 2
        inputs2 = Input(shape=(length,))
        embedding2 = Embedding(vocab_size, 100)(inputs2)
        conv2 = Conv1D(filters=32, kernel_size=6, activation='relu')(embedding2)
        drop2 = Dropout(0.5)(conv2)
        pool2 = MaxPooling1D(pool_size=2)(drop2)
        flat2 = Flatten()(pool2)
        # channel 3
        inputs3 = Input(shape=(length,))
        embedding3 = Embedding(vocab_size, 100)(inputs3)
        conv3 = Conv1D(filters=32, kernel_size=8, activation='relu')(embedding3)
        drop3 = Dropout(0.5)(conv3)
        pool3 = MaxPooling1D(pool_size=2)(drop3)
        flat3 = Flatten()(pool3)
        # merge
        merged = concatenate([flat1, flat2, flat3])
        # interpretation
        dense1 = Dense(10, activation='relu')(merged)
        outputs = Dense(1, activation='sigmoid')(dense1)
        model = Model(inputs=[inputs1, inputs2, inputs3], outputs=outputs)

        # compile
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        # summarize
        print(model.summary())
        #plot_model(model, show_shapes=True, to_file='multichannel.png')
        return model

if __name__ == '__main__':
    # load training dataset
    trainLines, trainLabels = sentimentModel().load_dataset('train.pkl')
    # create tokenizer
    tokenizer = sentimentModel().create_tokenizer(trainLines)
    print("Tokenize " + str(tokenizer.document_count))

    # calculate max document length
    length = sentimentModel().max_length(trainLines)
    # calculate vocabulary size
    vocab_size = len(tokenizer.word_index) + 1
    print('Max document length: %d' % length)
    print('Vocabulary size: %d' % vocab_size)
    # encode data
    trainX = sentimentModel().encode_text(tokenizer, trainLines, length)
    print(trainX.shape)

    # define model
    model = sentimentModel().define_model(length, vocab_size)
    # fit model
    model.fit([trainX, trainX, trainX], array(trainLabels), epochs=10, batch_size=16)
    # save the model
    model.save('model.h5')



import pyLDAvis.gensim
import pickle
import gensim

class ldaInference:
    def __init__(self, dictionary_model='dictionary.gensim', corpus_model='corpus.pkl', lda_model='model5.gensim'):
        self.dictionary = gensim.corpora.Dictionary.load(dictionary_model)
        self.corpus = pickle.load(open(corpus_model, 'rb'))
        self.lda = gensim.models.ldamodel.LdaModel.load(lda_model)

    def infer(self, document):
        test_doc = [self.dictionary.doc2bow(document.split(" "))]
        inferred_matrix = self.lda.inference(test_doc)

        return inferred_matrix

if __name__ == '__main__':
    a_document = '한국 시장경제가 위기입니다.'
    inferred_topics = ldaInference().infer(a_document)

    print(str(inferred_topics))



import math
import re
from decimal import Decimal


class NumUtil(object):
    """
    - number util
    """
    UK = 100000000  # 억

    @staticmethod
    def int2digit(n, base=10):  # base: 진수
        res = ''
        while n > 0:
            n, r = divmod(n, base)
            res = str(r) + res
        return res

    @staticmethod
    def comma_str(n, precision=0):
        """
        :param n: 수치 (문자열 또는 수치형)
        :param precision: 소수점 자리수
        """
        try:
            if n is None:
                return None
            if precision > 0:
                n = float(n)
            else:
                n = int(float(n))

            if isinstance(n, int):
                return r'{:,d}'.format(n)
            elif isinstance(n, float):
                rule = r'{:,.%df}' % precision
                return rule.format(n)
            else:
                return n
        except Exception:  # @UnusedVariable
            return n

    @staticmethod
    def auto_convert(num):
        try:
            return int(num)
        except:
            try:
                return float(num)
            except:
                try:
                    return str(num)
                except:
                    return num

    @staticmethod
    def to_digit(num):
        try:
            _num = ''
            for a in str(num):
                if ('0' <= a <= '9') or a in '-.':
                    _num += a
            _num = _num.lstrip('0')
            return int(_num)
        except:
            return 0

    @staticmethod
    def has_digit(num):
        for a in str(num):
            if a.isdigit():
                return True
        return False

    @staticmethod
    def remove_comma(line):
        return re.sub(r'''(\d),(\d{1})''', r'''\1\2''', line)

    @staticmethod
    def to_readable(n, min_decimal=0.001, max_decimal=1000):
        try:
            if (0 < math.fabs(n) < min_decimal) or (max_decimal < math.fabs(n)):
                return "{:.0e}".format(Decimal(n)).replace('+', '')  # scientific notation
            else:
                return n  # decimal notation
        except:
            return n

    @staticmethod
    def is_float(string):
        try:
            float(string)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_int(string):
        try:
            int(string)
            return True
        except ValueError:
            return False

    @staticmethod
    def base_num(v: float):
        try:
            v_str = str(float(v))
            if v_str.endswith('.0'):
                v_str = v_str.replace('.0', '')

            if v_str.count('.') > 0:
                v_str = v_str.split('.')[1]
                return 10 ** (-len(v_str))
            else:
                return 10 ** (v_str.count('0'))

            # if v_str.endswith('.0'):
            #     v_str = v_str.replace('.0', '')
            # # print()
            # # print(v_str)
            # try:
            #     a, b = v_str.split('.')
            #     # print('a:', a, 'b:', b)
            #     return 10 ** (-len(b))
            # except:
            #     return 10 ** (len(v_str)-1)
        except:
            return None


if __name__ == '__main__':
    print(NumUtil.base_num(3.5))
    print(NumUtil.base_num(4))
    print(NumUtil.base_num(0.001))
    print(NumUtil.base_num(0.018))
    print(NumUtil.base_num(0.00100))
    print(NumUtil.base_num(2120.00))
    print(NumUtil.base_num(700))
    # print(NumUtil.to_digit('-7,097,985.0원'))
    # print(NumUtil.has_digit('-a22a'))
    # print(type(NumUtil.auto_convert('20160101')))
    # print(type(NumUtil.auto_convert('2016-01-01')))
    # print(type(NumUtil.auto_convert('+1')))
    # print(type(NumUtil.auto_convert('-1')))
    # print(type(NumUtil.auto_convert('-1.1')))
    # print(NumUtil.to_digit('0001020'))
    # print(NumUtil.to_digit('0000000'))
    #    print(NumUtil.int2digit(8)
    #    print(NumUtil.int2digit(8, 2)
    #    print(NumUtil.int2digit(8, 16)
    #    print(NumUtil.comma_str(-100000)
    # print(NumUtil.remove_comma('123,456,789'))
    # print(NumUtil.remove_comma('브렉시트 충격서 벗어나는 코스피…나흘째 올라 1,970선 회복'))
    # print(NumUtil.remove_comma('가,나,다 123,456,789.00'))
    # print(NumUtil.remove_comma('23,456,789.00억원'))
    # print(NumUtil.to_readable(0.001))
    # print(NumUtil.to_readable(0.0001))
    # print(NumUtil.to_readable(1000))
    # print(NumUtil.to_readable(10000))
    # print(NumUtil.to_readable('title'))
    # print(NumUtil.comma_str(123456789012345.1234))
    # print(NumUtil.comma_str(123456789012345.1234, 2))
    # print(NumUtil.comma_str(123456789012345, 1))
    # print(NumUtil.comma_str('123456789012345.1234'))
    # print(NumUtil.comma_str('123456789012345.1234', 2))
    # print(NumUtil.comma_str('123456789012345', 1))
    pass


import numpy as np
from sklearn.preprocessing import LabelBinarizer


class CharOneHotVector(object):
    def __init__(self, chars: list, added: list = [], unkown_char=' '):
        chars.extend(added)
        if not chars or type(chars) is not list or len(chars) == 0:
            raise Exception('values must be list and len(values)>0 %s' % chars)

        self.unkown_char = unkown_char
        self.chars = chars
        self.encoder = LabelBinarizer(neg_label=0, pos_label=1, sparse_output=False)
        self.encoder.fit(chars)

    @property
    def classes(self):
        return self.encoder.classes_

    def __len__(self):
        return self.encoder.classes_.shape[0]

    @property
    def size(self):
        return self.encoder.classes_.shape[0]

    def __repr__(self):
        return '%s(len:%s)' % (self.__class__.__name__, self.__len__())

    def to_vector(self, char: str) -> np.ndarray:
        """

        :param char: character. len(c)==1
        :return:
        """
        return self.encoder.transform([char])[0]

    def to_vectors(self, chars: list, to_2dim=False) -> np.ndarray:
        """

        :param chars: list of characters. len(chars)>0
        :param to_1dim:
        :return:
        """
        if type(chars) is str or type(chars) is np.str_:
            chars = [c for c in chars]

        if to_2dim:
            return np.expand_dims(np.concatenate(self.encoder.transform(chars)), axis=0)
        else:
            return self.encoder.transform(chars)

    def to_value(self, vector: np.ndarray) -> np.ndarray:
        """

        :param vector: one hot vector
        :return:
        """
        if vector.ndim != 1:
            vector = vector.flatten()

        if not vector.any():
            return self.unkown_char
        else:
            return self.encoder.inverse_transform(np.array([vector]))[0]

    def to_values(self, vectors: np.ndarray) -> np.ndarray:
        """

        :param vectors: list of one hot vector
        :return:
        """
        if vectors.ndim != 2:
            vectors = vectors.reshape((len(vectors) // self.size, self.size))

        chars = []
        for vector in vectors:
            chars.append(self.to_value(vector))
        return ''.join(chars)
        # return ''.join(self.encoder.inverse_transform(vectors))

    def to_index(self, c: str) -> int:
        return np.argmax(self.to_vector(c))

    def index2value(self, index):
        if 0 < index < len(self.chars):
            return self.classes[index]
        else:
            return ''


if __name__ == '__main__':
    # chars = ['0', '1', '2']
    chars = ['ㄷ', 'ㄱ', 'ㄴ', 'ㄹ']
    chars = [0, 1]
    ohv = CharOneHotVector(chars)
    print(ohv.to_vectors([0, 1]))
    exit()

    _input = 'ㄱㄴㄷㄹㅎ'
    feature_v = ohv.to_vectors(_input)
    print(_input)
    print(feature_v)
    # print(ohv.to_value(feature_v[-1]))
    print(ohv.to_values(feature_v))
    print(ohv.to_values(np.array([[0.1, 1., 0.5, 0.5]])))

    # unary_vector = OneHotVector([0])
    # binary_vector = OneHotVector([0, 1])
    # ternary_vector = OneHotVector([0, 1, 2])
    # print(unary_vector, binary_vector, ternary_vector)
    # print('%6s\t%6s\t%6s\t%6s' % ('', 'unary', 'binary', 'ternary'))
    # for i in [0, 1, 2]:
    #     print('%6s\t%6s\t%6s\t%6s' % (i, unary_vector.to_vector(i), binary_vector.to_vector(i), ternary_vector.to_vector(i)))

    # @formatter:off
    # for c in chars:
    #     v = ohv.to_vector(c)
    #     print(c, type(v), v, ohv.to_value(v))
    #
    # vectors = ohv.to_vectors(chars)
    # print(type(vectors), vectors)
    # values = ohv.to_values(vectors)
    # print(type(values), values)


import re


class StringUtil(object):
    """
    - all about python string.
    """
    REGEX_TOKEN = r''' |=|\(|\)|\[|\]|<|>'''

    @staticmethod
    def rreplace(s, src, dest, cnt=1):
        return dest.join(s.rsplit(src, cnt))

    @staticmethod
    def mask_passwd_in_url(url):
        """
        e.g. mongodb://root:passwd@db-local:27017/... -> mongodb://root:******@db-local:27017/...
        :param url:
        :return:
        """
        try:
            a, b = url.index(r'://') + len(r'://'), url.index('@')
            head, auth, tail = url[:a], url[a:b], url[b:]
            if len(auth) > 0:
                user, passwd = auth.split(':')
                if len(passwd) > 0:
                    passwd = '*' * len(passwd)
                    auth = '%s:%s' % (user, passwd)
            return head + auth + tail
        except:
            return url

    @staticmethod
    def split_by_bracket(s, regex=REGEX_TOKEN):
        return [t for t in re.split(regex, s) if len(t) > 0]

    @staticmethod
    def from_set(_set):
        s = str(list(_set)).replace(' ', '')
        return s[1: len(s) - 1]

    @staticmethod
    def to_set(s, delim=','):
        return set(s.split(delim))

    @staticmethod
    def to_hex_list(s):
        return [c for c in map(ord, s)]

    # @classmethod
    # def to_hex(cls, char):
    #     return hex(ord(char))

    @staticmethod
    def to_hex_str_list(s):
        return ['{0:x}'.format(c) for c in map(ord, s)]

    @staticmethod
    def extract(s, start_s, end_s):
        if s and s.count(start_s) > 0 and s.count(end_s) > 0:
            start = s.find(start_s) + len(start_s)
            end = s.find(end_s, start)
            if start > 0 and end > 0:
                return s[start:end]
        return ''

    @staticmethod
    def split_by_size(s, size=100):
        lines = s.split('\n')
        result = ''
        for line in lines:
            if len(result) + len(line) > size:
                yield result
                result = line + '\n'
            else:
                result += line + '\n'
        yield result

    @staticmethod
    def merge(x, merge_tabs=False, merge_newlines=False):
        """
        TEXT인 경우, 공백, 탭, 줄바꿈 종류별로 merge를 한다.
        :param x:
        :param merge_tabs:
        :param merge_newlines:
        :return:
        """
        x = re.sub(' +', ' ', x)  # replace multiple blanks to one blank.
        if merge_tabs:
            x = re.sub(r'\t+', r'\t', x)
        if merge_newlines:
            x = re.sub(r'\n+', r'\n', x)
        return x.strip()

    @staticmethod
    def merge_to_one_line(x):
        """
        HTML인 경우, 다수의 공백, 탭, 줄바꿈을 유지할 필요가 없으므로, 모두 단일 공백으로 합쳐준다.
        :param x:
        :return:
        """
        return ' '.join(x.split()).strip()  # replace multiple blanks, tabs, new lines to one blank.

    @staticmethod
    def find_nth(s, x, n, i=0):
        i = s.find(x, i)
        if n == 1 or i == -1:
            return i
        else:
            return StringUtil.find_nth(s, x, n - 1, i + len(x))

    # @staticmethod
    # def split_and_recover(txt, delim):
    #     txt = txt.replace('\t', ' ')
    #     txt = txt.replace(delim, delim + '\t')
    #
    #     parts = txt.split('\t')
    #     return parts


    @staticmethod
    def remove_comment_line(text):
        lines = []
        for line in text.splitlines():
            if not line.startswith('#') and len(line) > 0:
                lines.append(line)
        return ' '.join(lines)

    @staticmethod
    def replace_with_index(s, newstring, index, nofail=False):
        if not nofail and index not in range(len(s)):
            raise ValueError("index outside given string")  # raise an error if index is outside of the string

        # if not erroring, but the index is still not in the correct range..
        if index < 0:  # add it to the beginning
            return newstring + s
        if index > len(s):  # add it to the end
            return s + newstring

        # insert the new string between "slices" of the original
        return s[:index] + newstring + s[index + 1:]


if __name__ == '__main__':
    print(StringUtil.replace_with_index('01234567890', 'ab', 4))
    # a = '\ufeffabc'
    # print('"%s"' % a.replace('\ufeff', ''))
    # print(StringUtil.mask_passwd_in_url('mongodb://root:passwd@db-local:27017/admin?authMechanism=MONGODB-CR'))
    # print(StringUtil.mask_passwd_in_url('mongodb://root@db-local:27017/admin?authMechanism=MONGODB-CR'))
    # print(StringUtil.mask_passwd_in_url('mongodb://db-local:27017/admin?authMechanism=MONGODB-CR'))
    # print(StringUtil.mask_passwd_in_url('db-local:27017/admin?authMechanism=MONGODB-CR'))
    # print(StringUtil.split_by_bracket('(서울=포커스뉴스) 송민순 전 외교통상부 장관의 회고록 파문이 일파만파 커지며 정치권의 논란이 가열되고 있다.'))
    # print(StringUtil.merge_to_one_line('081116 스브스 인기가요 롱넘버 캡쳐       어쩌다보니 유천이가 없다..\t\t진짜 롱넘버는 레전드인듯T.T 멋지다구ㅠㅠㅠㅠ 나중에 끝에 가면 홍어삼합도아닌 삼합화음?ㅋㅋㅋㅋ 진심 그거대박 나 그부분 너무 좋음 ㅠㅠㅠ'))
    # for line in StringUtil.split_by_size('1234\n56\n7890', size=5):
    #     print(line)
    # print(StringUtil.merge('a    b  c\t\td\n\ne', merge_tabs=True, merge_newlines=True))
    # s1 = set([])
    # s2 = set([1])
    # s3 = set([1, 2, 3])
    # print(s1, StringUtil.from_set(s1))
    # print(s2, StringUtil.from_set(s2))
    # print(s3, StringUtil.from_set(s3))
    # print(StringUtil.to_hex_list('123'))
    # print(' '.join(StringUtil.to_hex_str_list('123')))
    # print(StringUtil.extract(r"""fn\.link\('(.+?)'""", r"""$$fn.link('/cate/329', '_self', '카테전체보기_에어컨'); return false;"""))
    # print(StringUtil.phone_number('010-2747-6116'))
    # print(StringUtil.phone_number('010+2747_6116'))
    pass


"""
Takes document as input and performs chunking accoring to a pattern to return a list of Candidate Keywords.

Input :
    document - text file
Output :
    list of possible keyphrases

"""
import nltk
from nltk import word_tokenize
from nltk.chunk import RegexpParser
from treform.tokenizer import Komoran

def extract_candidate_keywords(words, tags):
    #Get the words in the document
    #words = word_tokenize(document)
    # Chunk first to get 'Candidate Keywords'
    #tags = nltk.pos_tag(words)

    chunkGram = r""" PHRASE: 
                        {(<JJ>* <NN.*>+ <IN>)? <JJ>* <NN.*>+}
                """

    chunkParser = RegexpParser(chunkGram)
    chunked = chunkParser.parse(tags)

    candidate_keywords = []
    for tree in chunked.subtrees():
        if tree.label() == 'PHRASE':
            candidate_keyword = ' '.join([x for x,y in tree.leaves()])
            candidate_keywords.append(candidate_keyword)

    candidate_keywords = [w for w in candidate_keywords if len(w) > 3 and  len(w.split(' ')) < 6]
    return candidate_keywords


def extract_candidate_keywords_for_training(document, language='ko'):
    if language == "ko":
        komoran = Komoran()
        tags = komoran(document)

        chunkGram = r""" 
                        PHRASE: {<N.*>*<Suffix>?}
                    """
    elif language == "en":
        # Get the words in the document
        words = word_tokenize(document)
        #Chunk first to get 'Candidate Keywords'
        tags = nltk.pos_tag(words)
        chunkGram = r"""
                        PHRASE: {(<JJ>* <NN.*>+ <IN>)? <JJ>* <NN.*>+}
                        """
    chunkParser = RegexpParser(chunkGram)
    chunked = chunkParser.parse(tags)

    candidate_keywords = []
    for tree in chunked.subtrees():
        if tree.label() == 'PHRASE':
            candidate_keyword = ' '.join([x for x, y in tree.leaves()])
            candidate_keywords.append(candidate_keyword)

    candidate_keywords = [w for w in candidate_keywords if len(w) > 3 and len(w.split(' ')) < 6]

    return candidate_keywords


import math
import numpy as np

def tf_ig(corpus,package):
	
	labelset = package ["labelset"]
	weights = package ["weights"]
	doclist = package ["doclist"]
	doclist = dict(doclist)
	dictlist = {}
	doclen = {}
	n = len(corpus)
	worddict = {}

	for i in corpus:
		labell = i["label"]
		docl = i["document"]
		if labell not in doclen:
			doclen[labell] = 0
		doclen[labell] += i["length"]
		for j in i["split_sentence"].split():
			# ctlist : label —— doc —— word —— frequency
			if labell not in dictlist:
				dictlist[labell] = {}
			if j not in dictlist[labell]:
				dictlist[labell][j] = 1
			else:
				dictlist[labell][j] += 1

			# doclist : label —— word ——　doc set
			if labell not in doclist:
				doclist[labell] = {}
			if j not in doclist[labell]:
				doclist[labell][j] = set()
			doclist[labell][j].add(docl)

	for labell in labelset:
		for word in doclist[labell]:
			if word not in worddict:
				worddict[word] = 0
			worddict[word] += len(doclist[labell][word])

	for labell in labelset:
		weights[labell] = {}
		for word in doclist[labell]:
			a_b = worddict[word]
			a = len(doclist[labell][word])
			b = a_b - a
			c_d = sum([(worddict[x]) for x in worddict.keys() if x!=word])
			c = sum([len(doclist[labell][x]) for x in (doclist[labell]) if x!=word])
			d = c_d - c

			weights[labell][word] = -1.0*( (a+c)*1.0/n )*math.log((a+c)*1.0/n,2) + (a*1.0/n)*math.log( a*1.0/(a+b),2 ) + (c*1.0/n)*math.log(c*1.0/(c+d),2)

	tf_ig = {}
	for labell in labelset:
		tf_ig[labell] = {}
		for word in dictlist[labell]:
			tf_ig[labell][word] = dictlist[labell][word]*1.0 /(doclen[labell]*1.0)
			if word in weights[labell]:
				tf_ig[labell][word] *= abs(weights[labell][word])


	package ["labelset"] = labelset
	package ["weights"] = weights
	package ["doclist"] = doclist

	return tf_ig



from treform.collector.colBasic import *

class ColDonga(ColBasic):
    def getInfo(params):
        return '동아일보 수집기', 0

    def getName(self):
        return "colDonga"

    def getSearchURL(self):
        return 'http://news.donga.com/search'

    def initParameter(self, params):
        startDate = params['startDate'][0]
        endDate = params['endDate'][0]
        if params.get('lastDate'): endDate = params['lastDate']
        startDate = self.reformatDate(startDate, '%04d%02d%02d')
        endDate = self.reformatDate(endDate, '%04d%02d%02d')
        return {'startDate':startDate, 'endDate':endDate}

    def makeParameter(self, page, query, d):
        return {'p':page*15 - 14,
            'query':query,
            'check_news':'1',
            'more':'1',
            'sorting':'1',
            'search_date':'5',
            'v1':d['startDate'],
            'v2':d['endDate'],
            'range':'1',
        }

    def selectList(self, soup):
        ret = []
        for el in soup.select('.txt a'):
            link = el['href']
            ret.append(link)
        return ret

    def selectArticle(self, soup):
        return (ColBasic.getTextFromElement(soup.select_one('h2.title')),
                ColBasic.cleanText(soup.find('meta', {'property': 'article:published_time'})['content']),
                ColBasic.getTextFromElement(soup.select_one('.article_txt')))




import treform as ptm
from treform.sentiment.MLSentimentManager import MachineLearningSentimentAnalyzer

sentiAnalyzer = MachineLearningSentimentAnalyzer()

model = sentiAnalyzer.load('sentiment.model')
vectorizer_model = sentiAnalyzer.loadVectorizer(model_name='senti_vectorizer.model')

docs = ['오늘은 세상이 참 아름답게 보이네요! 감사합니다',
        '우울한 날이면 언제나 부정적인 생각에 아프다...']
predictions = sentiAnalyzer.predict(docs, model, vectorizer_model)
for i, predicted in enumerate(predictions):
    print(predicted + ' '
                      '.0.0for ' + docs[i])


import pandas as pd
import numpy as np
import re
import nltk
from nltk.corpus import stopwords

from numpy import array
from keras.preprocessing.text import one_hot
from keras.preprocessing.sequence import pad_sequences
from keras.models import Sequential
from keras.layers.core import Activation, Dropout, Dense
from keras.layers import Flatten, Conv1D, LSTM
from keras.layers import GlobalMaxPooling1D
from keras.layers.embeddings import Embedding
from sklearn.model_selection import train_test_split
from keras.preprocessing.text import Tokenizer

import seaborn as sns
from numpy import array
from numpy import asarray
from numpy import zeros

import matplotlib.pyplot as plt

#Importing and Analyzing the Dataset
movie_reviews = pd.read_csv("data/IMDB Dataset.csv")
movie_reviews.isnull().values.any()

print(str(movie_reviews.shape))

print(str(movie_reviews.head()))

print(str(movie_reviews["review"][3]))

sns.countplot(x='sentiment', data=movie_reviews)

#Data Preprocessing
def preprocess_text(sen):
    # Removing html tags
    sentence = remove_tags(sen)

    # Remove punctuations and numbers
    sentence = re.sub('[^a-zA-Z]', ' ', sentence)

    # Single character removal
    sentence = re.sub(r"\s+[a-zA-Z]\s+", ' ', sentence)

    # Removing multiple spaces
    sentence = re.sub(r'\s+', ' ', sentence)

    return sentence

TAG_RE = re.compile(r'<[^>]+>')

def remove_tags(text):
    return TAG_RE.sub('', text)

X = []
sentences = list(movie_reviews['review'])
for sen in sentences:
    X.append(preprocess_text(sen))

y = movie_reviews['sentiment']

y = np.array(list(map(lambda x: 1 if x=="positive" else 0, y)))

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.20, random_state=42)

#Preparing the Embedding Layer
tokenizer = Tokenizer(num_words=5000)
tokenizer.fit_on_texts(X_train)

X_train = tokenizer.texts_to_sequences(X_train)
X_test = tokenizer.texts_to_sequences(X_test)

# Adding 1 because of reserved 0 index
vocab_size = len(tokenizer.word_index) + 1

maxlen = 100

X_train = pad_sequences(X_train, padding='post', maxlen=maxlen)
X_test = pad_sequences(X_test, padding='post', maxlen=maxlen)

#create our feature matrix
embeddings_dictionary = dict()
glove_file = open('../embeddings/glove.6B.100d.txt', encoding="utf8")

for line in glove_file:
    records = line.split()
    word = records[0]
    vector_dimensions = asarray(records[1:], dtype='float32')
    embeddings_dictionary [word] = vector_dimensions
glove_file.close()

embedding_matrix = zeros((vocab_size, 100))
for word, index in tokenizer.word_index.items():
    embedding_vector = embeddings_dictionary.get(word)
    if embedding_vector is not None:
        embedding_matrix[index] = embedding_vector

def train(mode):
    if (mode is 'simple'):
        #Text Classification with Simple Neural Network
        model = Sequential()
        embedding_layer = Embedding(vocab_size, 100, weights=[embedding_matrix], input_length=maxlen , trainable=False)
        model.add(embedding_layer)

        model.add(Flatten())
        model.add(Dense(1, activation='sigmoid'))

    elif (mode is 'cnn'):
        #Text Classification with a Convolutional Neural Network
        model = Sequential()

        embedding_layer = Embedding(vocab_size, 100, weights=[embedding_matrix], input_length=maxlen, trainable=False)
        model.add(embedding_layer)

        model.add(Conv1D(128, 5, activation='relu'))
        model.add(GlobalMaxPooling1D())
        model.add(Dense(1, activation='sigmoid'))

    elif (mode is 'lstm'):
        #Text Classification with Recurrent Neural Network (LSTM)
        model = Sequential()
        embedding_layer = Embedding(vocab_size, 100, weights=[embedding_matrix], input_length=maxlen, trainable=False)
        model.add(embedding_layer)
        model.add(LSTM(128))

        model.add(Dense(1, activation='sigmoid'))
    else:
        print('no matched deep learning algorithm')

    return model

model = train("lstm")
model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['acc'])

print(model.summary())

#train our model
history = model.fit(X_train, y_train, batch_size=128, epochs=6, verbose=1, validation_split=0.2)
model.save("lstm_model.h5")
#evaluate the performance of the mode
score = model.evaluate(X_test, y_test, verbose=1)

print("Test Score:", score[0])
print("Test Accuracy:", score[1])

plt.title('model accuracy')
plt.ylabel('accuracy')
plt.xlabel('epoch')
plt.legend(['train','test'], loc='upper left')
plt.show()

plt.plot(history.history['loss'])
plt.plot(history.history['val_loss'])

plt.title('model loss')
plt.ylabel('loss')
plt.xlabel('epoch')
plt.legend(['train','test'], loc='upper left')
plt.show()

#Making Predictions on Single Instance
instance = X[57]
print(instance)

instance = tokenizer.texts_to_sequences(instance)

flat_list = []
for sublist in instance:
    for item in sublist:
        flat_list.append(item)

flat_list = [flat_list]

instance = pad_sequences(flat_list, padding='post', maxlen=maxlen)

results = model.predict(instance)

print("predicted: " + str(results))



from numpy import *
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder

def one_hot(data):
	
	values = array(data)
	
	# integer encode
	label_encoder = LabelEncoder()
	integer_encoded = label_encoder.fit_transform(values)
	
	# binary encode
	onehot_encoder = OneHotEncoder(sparse = False)
	integer_encoded = integer_encoded.reshape(len(integer_encoded),1)
	onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
	
	return onehot_encoded


import logging
import multiprocessing
import os.path
import sys

from treform.utility.base_util import is_my_pc
from treform.utility.file_util import FileUtil

lock = multiprocessing.Lock()

class LogUtil(object):
    inited = False
    __default_logger_name = 'root'
    __log = None
    source_filepath = ''

    # ===============================================================================
    # default logging config
    # ===============================================================================
    #    level = logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL
    filename = ''
    datefmt = '%Y-%m-%d %H:%M:%S'
    format = '[%(asctime)s][%(levelname)5s] %(message)s'  # '[%(asctime)s] %(message)s'

    @classmethod
    def get_logger(cls, source_filepath=None, level=logging.DEBUG, format=format, datefmt=datefmt, console_mode=False, multiprocess=False):
        """
        get global logging object with multiprocessing-safe.
        :param multiprocess:
        :param source_filepath:
        :param console_mode: default: False
        :param datefmt:
        :param format:
        :param level: logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL(logging.FATAL
        :return: logging object
        """
        with lock:
            if cls.__log and cls.__log.hasHandlers() and cls.__log.level == level:
                sys.stderr.write('reuse old logger (id:%s, log_path:%s), level:%s\n' % (
                    id(cls.__log), cls.source_filepath, cls.__log.level))
                return cls.__log

            cls.source_filepath = source_filepath
            if not console_mode and is_my_pc():
                console_mode = True

            if source_filepath is None:
                console_mode = True
                log_path = None
            else:
                log_path = cls.__log_path_from_path(source_filepath)

            error_log_path = None
            if log_path:
                error_log_path = log_path.replace('.log', '.error.log')
                # if delete_old_files:
                #     if log_path:
                #         DirUtil.rmdirs(os.path.dirname(log_path))

                if log_path:  # create log dir
                    os.makedirs(os.path.dirname(log_path), exist_ok=True)
                if error_log_path:  # create log dir
                    os.makedirs(os.path.dirname(error_log_path), exist_ok=True)

            print('log_path:', log_path)
            print('error_log_path:', error_log_path)

            cls.__log = LogUtil.__create_logger(log_path, error_log_path, level, console_mode=console_mode,
                                                format=format, datefmt=datefmt, multiprocess=multiprocess)
            cls.__log.setLevel(level)
            cls.__log.info('log level: %s' % logging.getLevelName(level))

            LogUtil.basigConfig(level, format=format, datefmt=datefmt)
            logging.getLogger("gensim").setLevel(logging.DEBUG)
            logging.getLogger("multiprocessing").setLevel(logging.ERROR)
            logging.getLogger("urllib3").setLevel(logging.ERROR)
            logging.getLogger("requests").setLevel(logging.ERROR)
            logging.getLogger('elasticsearch').setLevel(logging.ERROR)
            logging.getLogger('elasticsearch.trace').setLevel(logging.ERROR)
            logging.getLogger("py2neo").setLevel(logging.ERROR)
            logging.getLogger("py2neo.batch").setLevel(logging.ERROR)
            logging.getLogger("py2neo.cypher").setLevel(logging.ERROR)
            logging.getLogger("httpstream").setLevel(logging.ERROR)
            logging.getLogger("sqlalchemy").setLevel(logging.ERROR)
            logging.getLogger("paramiko").setLevel(logging.INFO)
            # print('craete logger (%s, id=%s, log_path=%s), level=%s' % (cls.__log, id(cls.__log), log_path, level))

            return cls.__log

    @staticmethod
    def basigConfig(level=logging.NOTSET, format=format, datefmt=datefmt):
        logging.basicConfig(
            level=level,
            format=format,
            datefmt=datefmt
        )

    @staticmethod
    def __log_path_from_path(source_file, sub_log_dir='logs'):
        """
        :param sub_log_dir:
        :param source_file:
        :e.g.           ::LogUtil.get_logger(__file__) ./a.py -> ./logs/a.log
        :e.g.           ::LogUtil.get_logger(__file__, sub_log_dir='xx') ./a.py -> ./xx/a.log
        """
        _dir = os.path.join(os.path.dirname(str(source_file)), str(sub_log_dir))
        _basename = os.path.basename(str(source_file))
        if len(sys.argv) > 1:
            _basename = '%s.%s' % (_basename, FileUtil.to_filename('.'.join(sys.argv[1:])))
        log_path = os.path.join(_dir, _basename) + '.log'
        return log_path

    @staticmethod
    def __create_logger(log_path, error_log_path=None, level=logging.NOTSET, console_mode=False, format=format,
                        datefmt=datefmt, multiprocess=False):
        if not LogUtil.inited:
            LogUtil.inited = True
            # sys.stderr.write('console_mode: %s\n' % console_mode)
            # sys.stderr.write('level: %s\n' % logging.getLevelName(level))
            if log_path is not None:
                sys.stderr.write('log_path: %s\n' % log_path)
            if error_log_path is not None:
                sys.stderr.write('error_log_path: %s\n' % error_log_path)

        log = logging.getLogger(
            log_path)  # for single process but MultProcTimedRotatingFileHandler support multiprocessing
        # log = multiprocessing.get_logger()  # for multiple processes

        formatter = logging.Formatter(fmt=format, datefmt=datefmt)
        # ===============================================================================
        # logging handlers
        # ===============================================================================
        log.propagate = False
        log.handlers = []

        if console_mode:
            _consoleHandler = logging.StreamHandler(stream=sys.stdout)  # console logger
            _consoleHandler.setLevel(level)
            _consoleHandler.setFormatter(logging.Formatter(fmt='%(message)s', datefmt=datefmt))
            log.addHandler(_consoleHandler)

            _console_errorHandler = logging.StreamHandler(stream=sys.stderr)  # console logger
            _console_errorHandler.setLevel(logging.ERROR)
            _console_errorHandler.setFormatter(logging.Formatter(fmt='%(message)s', datefmt=datefmt))
            log.addHandler(_console_errorHandler)

        if log_path:
            if multiprocess:
                from .mult_proc_timed_rotating_file_handler import MultProcTimedRotatingFileHandler
                filehandler_class = MultProcTimedRotatingFileHandler
            else:
                from logging.handlers import TimedRotatingFileHandler
                filehandler_class = TimedRotatingFileHandler

            _fileHandler = filehandler_class(filename=log_path, when='midnight', interval=1, backupCount=0,
                                             encoding='utf8')
            _fileHandler.setLevel(level)
            _fileHandler.setFormatter(formatter)
            log.addHandler(_fileHandler)

            if error_log_path:
                _file_errorHandler = filehandler_class(filename=error_log_path, when='midnight', interval=1,
                                                       backupCount=0, encoding='utf8')
                _file_errorHandler.setLevel(logging.ERROR)
                _file_errorHandler.setFormatter(formatter)
                log.addHandler(_file_errorHandler)
        return log

    @classmethod
    def add_to_app_logger(cls, app):
        for h in cls.__log.handlers:
            app.logger.addHandler(h)
        return app.logger


if __name__ == '__main__':
    log = LogUtil.get_logger(source_filepath=None, level=logging.INFO, console_mode=True)
    log.debug('debug level OK.')
    log.info('info level OK.')
    log.error('error level OK.')


''' str => str '''

class BaseStemmer:
    IN_TYPE = [str]
    OUT_TYPE = [str]

class Porter(BaseStemmer):
    def __init__(self):
        import nltk
        self.inst = nltk.stem.PorterStemmer()

    def __call__(self, *args, **kwargs):
        return self.inst.stem(args[0])

class Lancaster(BaseStemmer):
    def __init__(self):
        import nltk
        self.inst = nltk.stem.LancasterStemmer()

    def __call__(self, *args, **kwargs):
        return self.inst.stem(args[0])


import random

from sklearn import metrics
from sklearn.metrics import confusion_matrix

import treform as ptm
from treform.sentiment.MLSentimentManager import MachineLearningSentimentAnalyzer

labels = ['pos','neg']

language = 'ko'
pipeline = None

_train_negative_docs = []
_test_negative_docs = []
_train_positive_docs = []
_test_positive_docs = []


def read_english_corpus():
    '''
    for English sentiment datasets
    '''

    #_train_negative_docs = ptm.CorpusFromDirectory('txt_sentoken/neg', True)
    #_train_negative_docs = _train_negative_docs.docs[:200]
    #_train_positive_docs = ptm.CorpusFromDirectory('txt_sentoken/pos', True)
    #_train_positive_docs = _train_positive_docs.docs[:200]
    #_test_negative_docs = ptm.CorpusFromDirectory('txt_sentoken/neg', False)
    #_test_negative_docs = _test_negative_docs.docs[:100]
    #_test_positive_docs = ptm.CorpusFromDirectory('txt_sentoken/pos', False)
    #_test_positive_docs = _test_positive_docs.docs[:100]

    _train_negative_docs = ptm.CorpusFromFile('./datasets/twitter/train-neg.txt')
    _train_negative_docs = _train_negative_docs.docs[:5000]
    _train_positive_docs = ptm.CorpusFromFile('./datasets/twitter/train-pos.txt')
    _train_positive_docs = _train_positive_docs.docs[:5000]
    _test_negative_docs = ptm.CorpusFromFile('./datasets/twitter/test-neg.txt')
    _test_negative_docs = _test_negative_docs.docs[:5000]
    _test_positive_docs = ptm.CorpusFromFile('./datasets/twitter/test-pos.txt')
    _test_positive_docs = _test_positive_docs.docs[:5000]

    return _train_negative_docs, _train_positive_docs, _test_negative_docs, _test_positive_docs

def read_korean_corpus():
    '''
    for Korean sentiment datasets
    '''

    corpus = ptm.CorpusFromFieldDelimitedFileForClassification('./datasets/ratings.txt', doc_index=1, class_index=2)
    combined_docs = corpus.docs[1:15000]

    positive_docs = []
    negative_docs = []
    labels = []

    idx = 1
    for doc in combined_docs:
        label = corpus.pair_map[idx]
        label = label.strip()
        labels.append(label)

        if int(label) == 0:
            negative_docs.append(doc)
        elif int(label) == 1:
            positive_docs.append(doc)
        idx += 1

    print(str(len(negative_docs)) + " : " + str(len(positive_docs)))

    neg_train_size = int((len(negative_docs) * 0.8) / len(negative_docs))
    _train_negative_docs = negative_docs[0:neg_train_size]
    _test_negative_docs = negative_docs[neg_train_size+1:len(negative_docs) - 1]

    pos_train_size = int((len(positive_docs) * 0.8) / len(positive_docs))
    _train_positive_docs = positive_docs[0:pos_train_size]
    _test_positive_docs = positive_docs[pos_train_size+1:len(positive_docs) - 1]

    return _train_negative_docs, _train_positive_docs, _test_negative_docs, _test_positive_docs, labels

if language == 'en':
    _train_negative_docs, _train_positive_docs, _test_negative_docs, _test_positive_docs \
                                                                         = read_english_corpus()
elif language == 'ko':
    _train_negative_docs, _train_positive_docs, _test_negative_docs, _test_positive_docs, labels \
                                                                         = read_korean_corpus()
if language == 'ko':
    mecab_path = 'C:\\mecab\\mecab-ko-dic'
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.Komoran(),
                            #ptm.tokenizer.MeCab(mecab_path),
                            ptm.helper.POSFilter('NN*|V*|IC*'),
                            ptm.helper.SelectWordOnly(),
                            ptm.ngram.NGramTokenizer(1, 2),
                            ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt')
                            )
elif language == 'en':
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.WordPos(),
                            ptm.helper.POSFilter('NN*|A*|V*|J*'),
                            ptm.helper.SelectWordOnly(),
                            #ptm.ngram.NGramTokenizer(1, 2),
                            ptm.helper.StopwordFilter(file='../stopwords/stopwordsEng.txt')
                            )

def make_documents(result):
    docs = []
    for doc in result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        docs.append(' '.join(new_doc))

    return docs

_train_negative = pipeline.processCorpus(_train_negative_docs)
train_negative = make_documents(_train_negative)
_train_positive = pipeline.processCorpus(_train_positive_docs)
train_positive = make_documents(_train_positive)

_test_negative = pipeline.processCorpus(_test_negative_docs)
test_negative = make_documents(_test_negative)
_test_positive = pipeline.processCorpus(_test_positive_docs)
test_positive = make_documents(_test_positive)

print('==  ==')

sentiAnalyzer = MachineLearningSentimentAnalyzer()
train_pos_vec = []
train_neg_vec = []
test_pos_vec = []
test_neg_vec = []
train_X = []
train_Y = []
test_X = []
test_Y = []

#training_method = customized, baseline
training_method = 'baseline'
if training_method == 'customized':
    train_pos_vec, train_neg_vec, test_pos_vec, test_neg_vec = \
                                    sentiAnalyzer.make_customized_feature_vectors(train_positive,
                                                                                  train_negative,
                                                                                  test_positive,
                                                                                  test_negative)
    features = train_pos_vec + train_neg_vec + test_pos_vec + test_neg_vec
    labels = ["neg"] * len(train_neg_vec) + ["pos"] * len(train_pos_vec) \
             + ["neg"] * len(test_neg_vec) + ["pos"] * len(test_pos_vec)

    train_X, test_X, train_Y, test_Y = \
                                    sentiAnalyzer.make_feature_vectors(features,
                                                                       labels)
elif training_method == 'baseline':

    combined_docs = train_negative + train_positive + test_negative + test_positive
    # random.shuffle(_combined_docs)
    labels = ["neg"] * len(train_negative) + ["pos"] * len(train_positive) + ["neg"] * len(test_negative) + ["pos"] * len(test_positive)

    #vectorizer_name is either count or tfidf (count -- countvectorizer or tfidf -- tfidfvectorizer)
    vectorizer_name = 'tfidf'
    if vectorizer_name == 'count':
        train_X, test_X, train_Y, test_Y = \
                                    sentiAnalyzer.make_baseline_feature_vectors(combined_docs,
                                                                                labels)
    elif vectorizer_name == 'tfidf':
        train_X, test_X, train_Y, test_Y = \
                                    sentiAnalyzer.make_tfidf_feature_vectors(combined_docs,
                                                                                labels)

customized = False
#algorithm is either lr, nb, svm, rf
algorithm = 'svm'

if customized == True:
    model = sentiAnalyzer.build_models_with_customizedvectors(train_pos_vec, train_neg_vec, algorithm)
    sentiAnalyzer.save(model)

    print("ML-based model")
    print("-----------")
    sentiAnalyzer.evaluate_model(model, test_pos_vec, test_neg_vec, True)
    print("")

else:
    model = sentiAnalyzer.build_baseline_model(train_X, train_Y, algorithm)
    sentiAnalyzer.save(model)

    print("ML-based model")
    print("-----------")
    model.fit(train_X, train_Y)
    y_pred = model.predict(test_X)
    conf_mat = confusion_matrix(test_Y, y_pred)
    print(conf_mat)
    print(metrics.classification_report(test_Y, y_pred,
                                        target_names=["neg", "pos"]))



import csv

from treform.splitter import *
from treform.tokenizer import *
from treform.stemmer import *
from treform.lemmatizer import *
from treform.tagger import *
from treform.helper import *
from treform.counter import *
from treform.chunker import *
from treform.cooccurrence import *
from treform.graphml import *
from treform.ngram import *
from treform.segmentation import *
from treform.keyword import *
from treform.noun_extractor import *
from treform.pmi import *
from treform.collector import *
from treform.utility import *
from treform.abbreviations import *
from treform.synonym import *
from treform.syntactic_parser import *
from treform.spelling import *
from treform.weighting import *

from os import listdir
from .version import __version__

import pandas as pd

class Pipeline:
    def __init__(self, *pipelines):
        self.pipeline = pipelines
        self.collapse = self.checkType(pipelines)
        pass

    def checkType(self, pipeline):
        if not pipeline: return
        if pipeline[0].IN_TYPE != [str]: raise Exception("%s requires '%s' as input, but receives 'str'" % (type(pipeline[0]).__name__, pipeline[0].IN_TYPE))
        collapse = [[]]
        curType = pipeline[0].OUT_TYPE
        for p, q in zip(pipeline[:-1], pipeline[1:]):
            qt = q.IN_TYPE
            pt = p.IN_TYPE
            if qt == curType[-len(qt):]:
                collapse.append(curType[:-len(qt)])
                curType = curType[:-len(qt)] + q.OUT_TYPE
                continue
            raise Exception("%s requires '%s' as input, but receives '%s' from %s" % (type(q).__name__, qt, pt, type(p).__name__))
        return collapse

    def processCorpus(self, corpus):
        ''' process corpus through pipeline '''

        def apply(p, a, inst):
            if not a:
                return p(inst)
            if a[0] == list:
                return [apply(p, a[1:], i) for i in inst]
            if a[0] == dict:
                return {k:apply(p, a[1:], v) for k, v in inst}

        results = []
        for d in corpus:
            inst = d
            for p, c in zip(self.pipeline, self.collapse):
                inst = apply(p, c, inst)

            #a = np.array(inst)
            #print("array shape " + str(len(a.shape)) + " : " + str(isinstance(inst[0], list)))
            results.append(inst)

            # if len is 2, it is array of array
            '''
            documents = []
            for doc in inst:
                document = ''
                for sent in doc:
                    document += " ".join(sent)
                documents.append(document)
            results.append(documents)
            '''

        return results

class Corpus:
    def __init__(self, textList):
        self.pair_map = {}
        self.docs = textList

    def __iter__(self):
        return self.docs.__iter__()

    def __len__(self):
        return self.docs.__len__()

class CorpusFromFile(Corpus):
    def __init__(self, file):
        self.docs = open(file, encoding='utf-8').readlines()

class CorpusFromEojiFile(Corpus):
    def __init__(self, file):
        import re
        emoji_pattern = re.compile("["
                                   u"\U00010000-\U0010FFFF"
                                   "]+", flags=re.UNICODE)
        array = []
        line_number = 0
        with open(file, encoding='utf-8') as original:
            for line in original.readlines():
                line_number += 1
                try:
                    after = emoji_pattern.sub(r'', line)
                    array.append(after)
                except IndexError:
                    print('line number', line_number, 'txt 파일에서 확인요망')
                    array.append()
        self.docs = array


class CorpusFromFieldDelimitedFile(Corpus):
    def __init__(self, file, index):
        array = []
        line_count=0
        with open(file, encoding='utf-8') as ins:
            for line in ins.readlines():
                inside=line.split('\t')
                line_count+=1
                try:
                    in_in=inside[index]
                except IndexError:
                    print(line_count,'번째 에러 확인, txt 파일의 확인요망')
                array.append(in_in)
        self.docs = array

class CorpusFromCSVFile(Corpus):

    def __init__(self, file, index):
        array = []
        line_count=0
        with open(file, encoding='utf-8') as ins:
            reader = csv.reader(ins)
            for row in reader:
                line_count+=1
                in_in = ''
                try:
                    in_in=row[index]
                except IndexError:
                    print(line_count,'번째 에러 확인, txt 파일의 확인요망')
                array.append(in_in)
        self.docs = array

class CorpusFromFieldDelimitedEmojiFile(Corpus):
    def __init__(self, file, index):
        array = []
        line_count=0
        import re
        emoji_pattern = re.compile("["
                                   u"\U00010000-\U0010FFFF"
                                   "]+", flags=re.UNICODE)
        with open(file, encoding='utf-8') as ins:
            for line in ins.readlines():

                inside=line.split('\t')
                line_count+=1
                after = ''
                try:
                    in_in = inside[index]
                    after = emoji_pattern.sub(r'', in_in)
                except IndexError:
                    print(line_count,'번째 에러 확인, txt 파일의 확인요망')

                if len(after) > 0:
                    array.append(after)

            self.docs = array

class CorpusFromFieldDelimitedFileWithYear(Corpus):
    def __init__(self, file, doc_index=1, year_index=0, delimiter='\t'):
        array = []
        id = 0
        pair_map = {}
        with open(file, encoding='utf-8') as ins:
            for line in ins:
                fields = line.split(delimiter)
                try:
                    array.append(fields[doc_index])
                    pair_map[id] = fields[year_index]

                    id += 1
                except IndexError:
                    print("out of index " + str(id))

        self.docs = array
        self.pair_map = pair_map


class CorpusFromFieldDelimitedFileForClassification(Corpus):
    def __init__(self, file, delimiter='\t',doc_index=1, class_index=0, title_index=-1):
        array = []
        id = 0
        pair_map = {}
        with open(file, encoding='utf-8') as ins:
            for line in ins:
                fields = line.split(delimiter)
                try:
                    doc = ''
                    if title_index != -1:
                        doc += ' ' + fields[title_index]

                    doc += ' ' + fields[doc_index]
                    array.append(doc.strip())
                    pair_map[id] = fields[class_index]

                    id += 1
                except IndexError:
                    print("out of index " + str(id))

        self.docs = array
        self.pair_map = pair_map

class CorpusFromDirectory(Corpus):
    def __init__(self, directory, is_train):
        array = []

        # walk through all files in the folder
        for filename in listdir(directory):
            # skip any reviews in the test set
            if is_train and filename.startswith('cv9'):
                continue
            if not is_train and not filename.startswith('cv9'):
                continue

            # create the full path of the file to open
            path = directory + '/' + filename
            # load the doc

            with open(path) as myfile:
                data = "".join(line.rstrip() for line in myfile)
                #print("data :: " + data)

            # add to list
            array.append(data)

        self.docs = array

class CorpusFromFieldDelimitedEmojiFile(Corpus):
    def __init__(self, file, index):
        import re
        emoji_pattern = re.compile("["
                                   u"\U00010000-\U0010FFFF"
                                   "]+", flags=re.UNICODE)
        array = []
        line_number = 0
        with open(file, encoding='utf-8') as original:
            for line in original.readlines():
                line_split = line.split('\t')
                line_number += 1
                try:
                    text = line_split[index]
                    after = emoji_pattern.sub(r'', text)
                    array.append(after)
                except IndexError:
                    print('line number', line_number, 'txt 파일에서 확인요망')
                    array.append()
        self.docs = array

class CorpusByDataFrame(Corpus):
    def __init__(self, file, delimiter, index, header=False):
        if header == False:
            df = pd.read_csv(file, delimiter=delimiter, header=None)
        else:
            df = pd.read_csv(file, delimiter=delimiter)
        selected = df.iloc[:,[index]]
        docs = []
        for s in selected.astype(str).values.tolist():
            docs.append(s[0])
        self.docs = docs




import math
import numpy as np

def iqf_qf_icf(corpus,package):
	weights = package["weights"]
	doclist = {}
	dictlist = {}
	worddict = {}
	word_label_dict = {}
	labelset = package["labelset"]

	n = len(corpus)

	for i in corpus:
		labell = i["label"]
		docl = i["document"]
		if labell not in doclist:
			doclist[labell] = {}
		if labell not in dictlist:
			dictlist[labell] = {}
		if docl not in dictlist[labell]:
			dictlist[labell][docl] = {}

		for j in i["split_sentence"].split():
			#dictlist : label —— doc —— word —— frequency
			if j not in dictlist[labell][docl]:
				dictlist[labell][docl][j] = 1
			else:
				dictlist[labell][docl][j] += 1
			#doclist : label —— word ——　doc set
			if j not in doclist[labell]:
				doclist[labell][j] = set()
			doclist[labell][j].add(docl)

	iqf_qf_icf = {}

	for labell in labelset:
		iqf_qf_icf[labell] = {}
		for word in doclist[labell]:
			if word not in word_label_dict:
				word_label_dict[word] = set()
			word_label_dict[word].add(labell)
			if word not in worddict:
				worddict[word] = 0
			worddict[word] += len(doclist[labell][word])

	#iqf_qf_icf
	for labell in labelset:
		weights[labell] = {}
		for word in doclist[labell]:
			a_b = worddict[word]
			a = len(doclist[labell][word])
			b = a_b - a
			weights[labell][word] = math.log(n*1.0/(a+b),2)*math.log(a+1,2)*math.log( len(labelset)*1.0/len(word_label_dict[word]) +1,2)
	
	package["weights"] = weights
	package["labelset"] = labelset

	return weights




import treform as ptm
import io
from nltk.corpus import sentiwordnet as swn
import nltk

class EnglishDictionarySentimentAnalyzer:
    def __init__(self):
        name = 'EnglishDictionarySentimentAnalyzer'

    def createDictionary(self):
        nltk.download('sentiwordnet')


if __name__ == '__main__':

    corpus = ptm.CorpusFromFile('../data/sampleEng.txt')
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            #ptm.tokenizer.Word(), #ptm.tokenizer.WordPos()
                            #ptm.helper.StopwordFilter(file='../stopwords/stopwordsEng.txt'),
                            #ptm.tagger.NLTK(),
                            ptm.tokenizer.WordPos(),
                            ptm.lemmatizer.WordNet())

    result = pipeline.processCorpus(corpus)

    EnglishDictionarySentimentAnalyzer().createDictionary()

    for doc in result:
        grand_score = 0.0
        count = 0
        for sent in doc:
            for _str in sent:
                _str[0]
                _str[1]
                pos = ''
                if (str(_str[1]).startswith("N")):
                    pos = 'n'
                elif (str(_str[1]).startswith("J")):
                    pos = 'a'
                elif (str(_str[1]).startswith("V")):
                    pos = 'v'
                try:
                    if (len(pos) > 0):
                        score = 0.0
                        breakdown = swn.senti_synset(str(_str[0]) + '.'+ pos + '.01')
                        #print(str(breakdown) + " " + str(breakdown.pos_score()) + " " + str(breakdown.neg_score()) + " " + str(breakdown.obj_score()))
                        if (breakdown.pos_score() > breakdown.neg_score()):
                            score = breakdown.pos_score()
                            count += 1
                        elif (breakdown.pos_score() < breakdown.neg_score() and breakdown.neg_score() != 0.0):
                            score = -breakdown.neg_score()
                            count += 1
                        grand_score += score
                except:
                    pos = ''

        avg_score = grand_score/count
        print("Average Sentiment Score " + str(avg_score))




import pickle
from sklearn.svm import LinearSVC
from sklearn.metrics import f1_score, confusion_matrix

def svm_train(X_train, X_test, y_train, y_test, C=1.0):
    #SVM training, optimal C=1.0, f1_score: 0.6834
    clf = LinearSVC(C=C, verbose=1, class_weight='balanced')
    clf.fit(X_train, y_train)
    pred = clf.predict(X_test)
    print("Train f1 %.3f, Test f1: %.3f" %(f1_score(y_train, clf.predict(X_train)), f1_score(y_test, pred, average='micro')))
    print("Test Confusion Matrix: ")
    print(confusion_matrix(y_test, pred))
    return clf


def main():
    X_train, X_test, y_train, y_test = pickle.load(open('word_mat_min.bin', 'rb'))
    print(X_train.shape, X_test.shape, y_train.shape, y_test.shape)

    #for i in [0.1, 0.5, 1.0, 2.0, 5.0]:
    #    svm_train(X_train, X_test, y_train, y_test, C=i)

    clf = svm_train(X_train, X_test, y_train, y_test)
    pickle.dump(clf, open('svm.model', 'wb'))

main()


import time

from treform.utility.date_util import DateUtil


class StopWatch:
    def __init__(self):
        self.Elapsed = 0.
        self.LastStart = None

    def start(self):
        assert self.LastStart is None
        self.LastStart = time.time()
        return self

    def stop(self):
        assert self.LastStart is not None
        self.Elapsed = time.time() - self.LastStart
        self.LastStart = None
        return self

    def elapsed(self):
        if self.LastStart is None:  # stopped
            return self.Elapsed
        else:
            return time.time() - self.LastStart


class WatchUtil(object):
    """
    - wrapper of 'pytools.stopwatch.StopWatch`
    - support multiple watches.
    """
    DEFAULT_WATCH_NAME = ''

    def __init__(self, auto_stop=True):
        """
        :param auto_stop: auto_stop=True이면, elapsed() 또는 elapsed_string() 호출시 자동으로 stop() 호출됨.
        """
        self.__watches = {}
        self.__cnt = {}
        self.auto_stop = auto_stop

    def __get(self, name=DEFAULT_WATCH_NAME) -> StopWatch:
        if name not in self.__watches:
            self.__watches[name] = StopWatch()
            self.__cnt[name] = 0
        return self.__watches[name]

    def del_watch(self, name=DEFAULT_WATCH_NAME) -> bool:
        if name in self.__watches:
            del self.__watches[name]
            return True
        else:
            return False

    def start(self, name=DEFAULT_WATCH_NAME) -> StopWatch:
        try:
            self.__get(name).stop()
        except:
            pass
        self.__get(name).start()
        self.__cnt[name] += 1
        return self.__get(name)

    def stop(self, name=DEFAULT_WATCH_NAME) -> StopWatch:
        return self.__get(name).stop()

    def elapsed(self, name=DEFAULT_WATCH_NAME) -> int:
        if self.auto_stop:
            try:
                self.__get(name).stop()  # must call start() later.
            except AssertionError:
                pass
        return self.__get(name).elapsed()

    def elapsed_string(self, name=DEFAULT_WATCH_NAME) -> str:
        return DateUtil.secs_to_string(self.elapsed(name))

    def summary(self, prefix='', include_total_time=False) -> str:
        import operator
        li = [(name, self.__watches[name].elapsed()) for name in self.__watches]
        li = sorted(li, key=operator.itemgetter(1), reverse=True)
        s = ''
        # for name, total_milli_secs in li:
        #     s += 'total [%s] %s\n' % (DateUtil.millisecs_to_string(total_milli_secs), name)
        if include_total_time:
            for name, total_milli_secs in li:
                if self.__cnt[name] > 0:
                    if len(prefix) > 0:
                        s += '%s total [%s] %s\n' % (
                            prefix, DateUtil.millisecs_to_string(float(total_milli_secs)), name)
                    else:
                        s += 'total [%s] %s\n' % (
                            DateUtil.millisecs_to_string(float(total_milli_secs)), name)

        for name, total_milli_secs in li:
            if self.__cnt[name] > 0:
                print(total_milli_secs)
                if len(prefix) > 0:
                    s += '%s average [%s] %s\n' % (
                        prefix, DateUtil.millisecs_to_string(float(total_milli_secs) / float(self.__cnt[name])), name)
                else:
                    s += 'average [%s] %s\n' % (
                        DateUtil.millisecs_to_string(float(total_milli_secs) / float(self.__cnt[name])), name)
        return s


if __name__ == '__main__':
    watches = WatchUtil(auto_stop=False)

    # watches.start('A')
    # time.sleep(1)
    # print(watches.elapsed_string('A'))
    # watches.del_watch('A')

    watches.start()
    time.sleep(3)
    print(watches.elapsed_string())

    watches.start()
    time.sleep(1)
    print(watches.elapsed_string())
    print(watches.elapsed_string())

    watches.start()
    time.sleep(3)
    print(watches.elapsed_string())

    print(watches.summary())



import tomotopy as tp
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as clr

class ExpNormalize(clr.Normalize):
    def __init__(self, scale):
        super().__init__()
        self.scale = scale

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        result, is_scalar = self.process_value(value)

        self.autoscale_None(result)
        (vmin,), _ = self.process_value(self.vmin)
        (vmax,), _ = self.process_value(self.vmax)
        if vmin == vmax:
            result.fill(0)
        elif vmin > vmax:
            raise ValueError("minvalue must be less than or equal to maxvalue")
        else:
            if clip:
                mask = np.ma.getmask(result)
                result = np.ma.array(np.clip(result.filled(vmax), vmin, vmax),
                                    mask=mask)
            resdat = result.data
            resdat = 1 - np.exp(-2 * resdat / self.scale)
            result = np.ma.array(resdat, mask=result.mask, copy=False)
        if is_scalar:
            result = result[0]
        return result

heat = clr.LinearSegmentedColormap.from_list('heat',
    [(0, 0, 0), (0, 0, 1), (0, 1, 1), (0, 1, 0), (1, 1, 0), (1, 0, 0), (1, 1, 1)],
    N=1024
)

corpus = tp.utils.Corpus()
for line in open('./topic_model/dataset2.txt', encoding='utf-8'):
    fd = line.strip().split()
    corpus.add_doc(fd[2:], metadata=list(map(float, fd[:2])))

# We set a range of the first metadata as [2000, 2017]
# and one of the second metadata as [0, 1].
mdl = tp.GDMRModel(tw=tp.TermWeight.PMI, k=10, degrees=[4, 3],
    alpha=1e-2, sigma=0.25, sigma0=3.0,
    metadata_range=[(2000, 2017), (0, 1)], corpus=corpus
)
mdl.optim_interval = 20
mdl.burn_in = 200

mdl.train(0)

print('Num docs:{}, Num Vocabs:{}, Total Words:{}'.format(
    len(mdl.docs), len(mdl.used_vocabs), mdl.num_words
))

# Let's train the model
for i in range(0, 1000, 20):
    print('Iteration: {:04} LL per word: {:.4}'.format(i, mdl.ll_per_word))
    mdl.train(20)
print('Iteration: {:04} LL per word: {:.4}'.format(1000, mdl.ll_per_word))

# Let's visualize the result
topic_counts = mdl.get_count_by_topics()
lambdas = mdl.lambdas

md_range = mdl.metadata_range
# Our topic distribution map has
# 400 pixels for the first axis and
# 200 pixels for the second axis.
r = mdl.tdf_linspace(
    [md_range[0][0], md_range[1][0]],
    [md_range[0][1], md_range[1][1]],
    [400, 200]
)

for k in (-topic_counts).argsort():
    print('Topic #{} ({})'.format(k, topic_counts[k]))
    print(*(w for w, _ in mdl.get_topic_words(k)))
    print('Lambda:', lambdas[k])

    imgplot = plt.imshow(r[:, :, k].transpose(), clim=(0.0, r[:, :, k].max()),
        origin='lower', cmap=heat, norm=ExpNormalize(scale=0.04),
        extent=[*md_range[0], *md_range[1]],
        aspect='auto'
    )
    plt.title('#{}\n({})'.format(k, ' '.join(w for w, _ in mdl.get_topic_words(k, top_n=5))))
    plt.colorbar()
    plt.show()


import json

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.feature_selection import chi2
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import LinearSVC
from sklearn.model_selection import cross_val_score
from sklearn.metrics import confusion_matrix
import seaborn as sns
from sklearn import metrics
from joblib import dump, load
from sklearn.neighbors import KNeighborsClassifier

import sklearn

class documentClassifier:
    def __init__(self):
        name='document_classifier'
        self.documents=[]
        self.features=None
        self.labels=None
        self.category_id_df=None
        self.df=None
        self.tfidf=None
        self.id_to_category=None
        self.category_to_id=None

        self.features=None
        self.labels=None
        self.class_number=0

    def preprocess(self, documents, class_list):

        #1. we assume that pyTextMiner pre-processing module was applied to the list of documents
        print(str(len(documents)) + " : " + str(len(class_list)))
        #create a list of rows consisting of text and class label by DataFrame
        self.df = pd.DataFrame({'text': documents, 'label': class_list})
        self.df = sklearn.utils.shuffle(self.df, n_samples=500, random_state=100)
        self.df.reset_index(inplace=True, drop=True)

        # Remove missing values in “text” column, and add a column encoding the label as an integer
        # because categorical variables are often better represented by integers than strings.
        # Create a couple of dictionaries for future use.
        # After cleaning up, this is the first five rows of the data we will be working on
        col = ['text', 'label']
        self.df = self.df[col]
        self.df = self.df[pd.notnull(self.df['text'])]
        self.df.columns = ['text', 'label']
        self.df['category_id'] = self.df['label'].factorize()[0]
        self.category_id_df = self.df[['label', 'category_id']].drop_duplicates().sort_values('category_id')
        self.category_to_id = dict(self.category_id_df.values)
        self.id_to_category = dict(self.category_id_df[['category_id', 'label']].values)

        # To save the dictionary into a file:
        json.dump(self.id_to_category, open("./model/id_to_category.json", 'w'))

        print(self.df.head())

        #number of classes
        self.class_number=self.df.label.unique()

        #visualize the distribution of classes
        fig = plt.figure(figsize=(8,6))
        self.df.groupby('label').text.count().plot.bar(ylim=0)
        plt.show()

        #2. vectorization
        self.tfidf = TfidfVectorizer(sublinear_tf=True, min_df=5, norm='l2', encoding='utf-8')
        self.features = self.tfidf.fit_transform(self.df.text).toarray()
        self.labels = self.df.category_id
        print(self.features.shape)

        #3. feature selection
        # We can use sklearn.feature_selection.chi2 to find the terms
        # that are the most correlated with each of the class labels
        N = 2
        for label, category_id in sorted(self.category_to_id.items()):
            features_chi2 = chi2(self.features, self.labels == category_id)
            indices = np.argsort(features_chi2[0])
            feature_names = np.array(self.tfidf.get_feature_names())[indices]
            unigrams = [v for v in feature_names if len(v.split(' ')) == 1]
            bigrams = [v for v in feature_names if len(v.split(' ')) == 2]
            print("# '{}':".format(label))
            print("  . Most correlated unigrams:\n. {}".format('\n. '.join(unigrams[-N:])))
            print("  . Most correlated bigrams:\n. {}".format('\n. '.join(bigrams[-N:])))

    def train(self, model_index=1):
        #4. supervised-based classification model selection
        # We are now ready to experiment with different machine learning models,
        # evaluate their accuracy and find the source of any potential issues.
        # We will benchmark the following four models:
        # Logistic Regression
        # (Multinomial) Naive Bayes
        # Linear Support Vector Machine
        # Random Forest
        models = [
            RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0),
            LinearSVC(),
            MultinomialNB(),
            LogisticRegression(random_state=0),
            KNeighborsClassifier(n_neighbors=10, weights='uniform', algorithm='auto', leaf_size=30, p=2, metric='minkowski', metric_params=None, n_jobs=None),
            SGDClassifier(loss='hinge', penalty='l2', alpha=0.0001)
        ]

        CV = 5
        cv_df = pd.DataFrame(index=range(CV * len(models)))
        entries = []
        for model in models:
            model_name = model.__class__.__name__
            accuracies = cross_val_score(model, self.features, self.labels, scoring='accuracy', cv=CV)
            for fold_idx, accuracy in enumerate(accuracies):
                entries.append((model_name, fold_idx, accuracy))
        cv_df = pd.DataFrame(entries, columns=['model_name', 'fold_idx', 'accuracy'])

        sns.boxplot(x='model_name', y='accuracy', data=cv_df)
        sns.stripplot(x='model_name', y='accuracy', data=cv_df,
                      size=8, jitter=True, edgecolor="gray", linewidth=2)
        plt.show()

        cv_df.groupby('model_name').accuracy.mean()

        #5. data preparation for training and validating
        # Continue with our best model (LinearSVC), we are going to look at the confusion matrix,
        # and show the discrepancies between predicted and actual labels.
        model = models[model_index]

        X_train, X_test, y_train, y_test, indices_train, indices_test = train_test_split(self.features, self.labels, self.df.index,
                                                                                         test_size=0.33, random_state=0)
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)

        return X_train, X_test, y_train, y_test, y_pred, indices_test, model

    def evaluate(self, y_test, y_pred, indices_test, model):
        #6. evaluation
        conf_mat = confusion_matrix(y_test, y_pred)
        fig, ax = plt.subplots(figsize=(10, 10))
        sns.heatmap(conf_mat, annot=True, fmt='d',
                    xticklabels=self.category_id_df.label.values, yticklabels=self.category_id_df.label.values)
        plt.ylabel('Actual')
        plt.xlabel('Predicted')
        plt.show()

        # The vast majority of the predictions end up on the diagonal (predicted label = actual label),
        # where we want them to be. However, there are a number of misclassifications,
        # and it might be interesting to see what those are caused by:
        for predicted in self.category_id_df.category_id:
            for actual in self.category_id_df.category_id:
                if predicted != actual and conf_mat[actual, predicted] >= 10:
                    print("'{}' predicted as '{}' : {} examples.".format(self.id_to_category[actual], self.id_to_category[predicted],
                                                                         conf_mat[actual, predicted]))
                    print(self.df.loc[indices_test[(y_test == actual) & (y_pred == predicted)]][
                                ['label', 'text']])
                    print('')

        # As you can see, some of the misclassified complaints are complaints that touch on more than one subjects
        # (for example, complaints involving both credit card and credit report). This sort of errors will always happen.
        # Again, we use the chi-squared test to find the terms that are the most correlated with each of the categories:
        model.fit(self.features, self.labels)
        N = 2
        if 'kneighbors' not in model.__class__.__name__.lower():
            for label, category_id in sorted(self.category_to_id.items()):
                indices = np.argsort(model.coef_[category_id])
                feature_names = np.array(self.tfidf.get_feature_names())[indices]
                unigrams = [v for v in reversed(feature_names) if len(v.split(' ')) == 1][:N]
                bigrams = [v for v in reversed(feature_names) if len(v.split(' ')) == 2][:N]
                print("# '{}':".format(label))
                print("  . Top unigrams:\n       . {}".format('\n       . '.join(unigrams)))
                print("  . Top bigrams:\n       . {}".format('\n       . '.join(bigrams)))



        #8. final evaluation per class
        print(metrics.classification_report(y_test, y_pred,
                                            target_names=self.df['label'].unique()))

    def save(self, model, model_name='classification.model'):
        dump(model, model_name)

    def saveVectorizer(self, model_name='vectorizer.model'):
        dump(self.tfidf, model_name)

    def load(self, model_name):
        return load(model_name)

    def loadVectorizer(self, model_name='vectorizer.model'):
        return load(model_name)

    def predict(self, model, vectorizer_model):
        #7. prediction
        docs = ["한국 경제 글로벌 위기 수요 위축 시장 경제 붕귀 자동차 수출 빨간불 내수 촉진 증진 방향성 제고",
                "밝기 5등급 정도 도심 밖 맨눈 충분히 관측 가능 새해 미국인 8월 행운 기대",
                "최순실 민간인 국정농단 의혹 사건 진상규명 국정조사 특별위원회 1차 청문회 이재용 삼성전자 부회장 재벌 총수 9명 증인 출석"]

        with open("./model/id_to_category.json") as handle:
            id_to_category = json.loads(handle.read())

        text_features = vectorizer_model.transform(docs)
        predictions = model.predict(text_features)
        for text, predicted in zip(docs, predictions):
            print('"{}"'.format(text))
            print("  - Predicted as: '{}'".format(id_to_category[str(predicted)]))
            print("")

    def predict_realtime(self, model, vectorizer_model, docs):

        with open("./model/id_to_category.json") as handle:
            id_to_category = json.loads(handle.read())

        text_features = vectorizer_model.transform(docs)
        predictions = model.predict(text_features)
        for text, predicted in zip(docs, predictions):
            print('"{}"'.format(text))
            print("  - Predicted as: '{}'".format(id_to_category[str(predicted)]))
            print("")


import math
import numpy as np

def tf_chi(corpus,package):
	labelset = package ["labelset"]
	weights = package ["weights"]
	doclist = package ["doclist"]
	doclist = dict(doclist)
	dictlist = {}
	doclen = {}
	totaldoc = len(corpus)
	worddict = {}

	for i in corpus:
		labell = i["label"]
		docl = i["document"]
		doclen[docl] = i["length"]
		for j in i["split_sentence"].split():

			if labell not in dictlist:
				dictlist[labell] = {}
			if docl not in dictlist[labell]:
				dictlist[labell][docl] = {}
			if j not in dictlist[labell][docl]:
				dictlist[labell][docl][j] = 1
			else:
				dictlist[labell][docl][j] += 1

			# doclist : label —— word ——　doc set
			if labell not in doclist:
				doclist[labell] = {}
			if j not in doclist[labell]:
				doclist[labell][j] = set()
			doclist[labell][j].add(docl)

	for labell in labelset:
		weights[labell] = {}
		for word in doclist[labell]:
			if word not in worddict:
				worddict[word] = 0
			worddict[word] += len(doclist[labell][word])

	# chi-square
	for labell in labelset:
		for word in doclist[labell]:
			a_b = worddict[word]
			a = len(doclist[labell][word])
			b = a_b - a
			c_d = sum([(worddict[x]) for x in worddict.keys() if x!=word])
			c = sum([len(doclist[labell][x]) for x in (doclist[labell]) if x!=word])
			d = c_d - c

			if (a_b*c_d*(b+d)*(a+c)) != 0:
				weights[labell][word] = totaldoc*1.0* (a*d-b*c)* (a*d-b*c) /(a_b*c_d*(b+d)*(a+c))

	#print chi
	# tf-chi
	tf_chi = {}
	for labell in labelset:
		tf_chi[labell] = {}
		for doc in dictlist[labell]:
			tf_chi[labell][doc] = {}
			for word in dictlist[labell][doc]:
				tf_chi[labell][doc][word] = dictlist[labell][doc][word]*1.0 / doclen[doc]
				if word in weights[labell]:
					tf_chi[labell][doc][word] *= weights[labell][word]

	package ["labelset"] = labelset
	package ["weights"] = weights
	package ["doclist"] = doclist

	return tf_chi
